import sys
import subprocess
from shutil import which


class Dependencies:

    def __init__(self):
        """
        Initialize the Dependencies class.

        :raises SystemExit: If any of the dependencies are not found in the system path.
        """
        self.dependencies = ["minimap2", "paf2gfa", "miniasm"]
        self.check_dependencies()


    def check_dependencies(self) -> None:
        """
        Check if all dependencies are present in the system path.

        :raises SystemExit: If any of the dependencies are not found in the system path.
        """
        for dependency in self.dependencies:
            setattr(self, dependency, find_exe(dependency))
            # logging.info(f'{dependency}: {getattr(self, dependency)}')
            if not getattr(self, dependency):
                sys.exit(f"Dependency {dependency} not found in path")



def find_exe(name: str) -> str | None:
    """
    Find the executable file with the specified name.

    :param name: The name of the executable.
    :return: The path to the executable, or None if not found.
    """
    # shutil.which seems to work mostly but is still not completely portable
    exe = which(name, path='/'.join(sys.executable.split('/')[:-1]))
    if not exe:
        exe = which(name)
    if not exe:
        exe = subprocess.run(f'which {name}', shell=True, capture_output=True, universal_newlines=True).stdout
    if not exe:
        return
    return exe.strip()




import logging
import time
import os

from boss.core import Boss
from boss.live import LiveRun
from boss.batch import FastqBatch
from boss.utils import execute
from boss.aeons.sequences import SequencePool, SequenceAVA, UnitigPool, ContigPool, Benefit
from boss.aeons.repeats import RepeatFilter



class BossAeons(Boss):


    def init(self):
        """
        Initialisation function to be run after the init of the superclass
        :return:
        """
        # initialise central objects
        self.pool = SequencePool(name=self.args.name,
                                 min_len=self.args.min_seq_len,
                                 out_dir=self.out_dir)
        self.ava = SequenceAVA(paf=f'{self.args.name}.ava',
                               tetra=self.args.tetra,
                               filters=self.args)
        # init scoring func
        self.score_vec = Benefit.init_scoring_vec(lowcov=self.args.lowcov)
        # launch into first assembly
        if hasattr(self.args, "live_run") and getattr(self.args, "live_run"):
            self.first_live_asm()



    def first_live_asm(self) -> None:
        """
        Construct a first assembly from some initial data.
        Scan directory for data and wait until args.data_wait megabases accumulate
        Loop until some contigs have been generated

        :return:
        """
        while True:
            new_fastq = LiveRun.scan_dir(fastq_pass=self.args.fq, processed_files=set())
            fq_batch = FastqBatch(fq_files=new_fastq, channels=self.channels)
            logging.info(f"available so far: {fq_batch.total_bases / 1e6} Mbases")
            if fq_batch.total_bases / 1e6 < self.args.data_wait:
                logging.info(f"waiting for {self.args.data_wait} Mbases of data ... \n")
                time.sleep(30)
                continue
            else:
                # try first asm
                logging.info("Attempting initial assembly")
                init_pool = SequencePool(name="init_pool",
                                         min_len=self.args.min_seq_len,
                                         out_dir=self.out_dir)
                init_pool.ingest(seqs=fq_batch.read_sequences)

                init_contigs = init_pool.initial_asm_miniasm()
                ncontigs = len(init_contigs.sequences)
                has_contig = init_pool.has_min_one_contig(min_contig_len=self.args.min_contig_len)

                if not ncontigs or not has_contig:
                    logging.info(f"Initial assembly yielded no contigs, waiting for more data ... ")
                    time.sleep(30)
                    continue
                else:
                    self.pool = SequencePool(name=self.args.name,
                                             min_len=self.args.min_seq_len,
                                             out_dir=self.out_dir)
                    self.ava = SequenceAVA(paf=f'{self.args.name}.ava',
                                           tetra=self.args.tetra,
                                           filters=self.args)
                    self.pool.ingest(init_contigs)
                    # set up a repeat filter from the initial set of reads
                    if self.args.filter_repeats:
                        self.repeat_filter = RepeatFilter(name=self.args.name, seqpool=init_pool)
                    break

        # once there are contigs, record used files
        self.processed_files = set()
        self.processed_files.update(new_fastq)
        self.n_fastq = len(new_fastq)
        logging.info("Initial asm completed\n\n")



    def assemble_unitigs(self) -> SequencePool:
        """
        Wrapper to construct new contigs from the current data.
        Removes used sequences in the process

        :return: SequencePool of the new unitigs
        """
        # write current links to paf
        self.ava.links2paf(paf_out=self.ava.paf_links)
        # write pool to file
        SequencePool.write_seq_dict(seq_dict=self.pool.seqdict(), file=self.pool.fa)
        # create gfa and unitigs
        gfa = self.ava.paf2gfa_gfatools(paf=self.ava.paf_links, fa=self.pool.fa)
        # load the new unitigs
        unitigs = SequencePool.load_unitigs(gfa=gfa)
        # put them into a collection
        unitig_pool = UnitigPool(unitigs)
        # get the coverage arrays
        unitig_pool.get_unitig_coverage_arrays(seqpool=self.pool)
        # transform into a sequence pool
        new_pool, used_sids = unitig_pool.unitigs2seqpool(
            seqpool=self.pool, min_seq_len=self.args.min_seq_len
        )
        # remove used sequences from current pool
        self.remove_seqs(used_sids)
        return new_pool


    def assemble_add_and_filter_contigs(self) -> SequencePool:
        """
        Main wrapper to update the current contigs.
        Assembles the current graph and extracts new unitigs,
        adds them to the seqpool and extracts contigs for mapping against

        :return: SequencePool of current contigs (old and new)
        """
        logging.info("assembling new unitigs.. ")
        new_pool = self.assemble_unitigs()
        # add new sequences to the dict and to the ava
        logging.info("loading and overlapping new unitigs.. ")
        self.add_new_sequences(sequences=new_pool, increment=False)
        # write the current pool to file for mapping against
        logging.info("finding contigs to map against.. ")
        contigs = self.pool.declare_contigs(min_contig_len=self.args.min_contig_len)
        SequencePool.write_seq_dict(seq_dict=contigs.seqdict(), file=self.pool.contig_fa)
        return contigs




    def remove_seqs(self, sequences: set[str]) -> None:
        """
        Wrapper to remove sequences from pool, ava, coverage etc.

        :param sequences: Set of read IDs
        :return:
        """
        if not sequences:
            return
        self.ava.remove_links(sequences=sequences)
        self.pool.remove_sequences(sequences=sequences)



    def add_new_sequences(self, sequences: SequencePool, increment: bool = True) -> None:
        """
        Wrapper to add new sequences to the main pool

        :param sequences: Sequences to add
        :param increment: Whether to increment coverage when adding new sequences
        :return:
        """
        logging.info("Adding new sequences")
        ava_new, ava_onto_pool = self.pool.add2ava(sequences)
        self.pool.ingest(seqs=sequences)
        # load new alignments
        cont_new, ovl_new = self.ava.load_ava(ava_new, seqpool=self.pool)
        if increment:
            self.pool.increment(containment=cont_new)
        cont_onto, ovl_onto = self.ava.load_ava(ava_onto_pool, seqpool=self.pool)
        if increment:
            self.pool.increment(containment=cont_onto)
        # remove contained sequences
        cont = SequenceAVA.source_union(edges0=cont_new, edges1=cont_onto)
        self.remove_seqs(sequences=cont)
        # raise temp for overlappers (new links are saved as class attr in load_ava)
        ovl = ovl_new | ovl_onto
        self.pool.reset_temperature(ovl, t=self.args.temperature)



    def overlap_pool(self) -> None:
        """
        Wrapper to run AVA for pool to find overlaps and remove contained sequences

        :return:
        """
        logging.info("Running all-versus-all of sequence pool")
        contigs = self.pool.declare_contigs(min_contig_len=self.args.min_contig_len)
        if contigs.is_empty():
            return
        pool_paf = self.pool.run_ava(sequences=contigs.seqdict(), fa=self.pool.fa, paf=self.pool.ava)
        pool_contained, pool_ovl = self.ava.load_ava(paf=pool_paf, seqpool=self.pool)
        self.pool.increment(containment=pool_contained)
        cont = SequenceAVA.source_union(edges0=pool_contained, edges1={})
        if cont:
            logging.info(f'Removing {len(cont)} contained sequences from pool')
            self.remove_seqs(sequences=cont)
        self.pool.reset_temperature(pool_ovl)



    def trim_sequences(self) -> None:
        """
        Wrapper to find reads that need trimming for potential overlaps
        done after load_ava where mappings are marked with c=6
        i.e. only trim stuff already in the pool

        :return:
        """
        logging.info('')
        trim_dict = self.ava.to_be_trimmed()
        logging.info(f"Trimming {len(trim_dict.keys())} sequences")
        # trim and ingest
        trimmed_seqs = self.pool.trim_sequences(trim_dict=trim_dict)
        trim_paf = self.pool.run_ava(sequences=trimmed_seqs,
                                     fa=f'{self.pool.fa}.trim',
                                     paf=f'{self.pool.ava}.trim')
        trim_contained, _ = self.ava.load_ava(paf=trim_paf, seqpool=self.pool)
        to_remove = self.ava.trim_success(trim_dict=trim_dict, overlaps=self.ava.overlaps)
        # remove original sequences & failed mergers
        self.remove_seqs(sequences=to_remove)





    def cleanup(self) -> None:
        """
        Move temp files after a simulation

        :return:
        """
        tmpdir = f'{self.out_dir}/tmp'
        if not os.path.exists(tmpdir):
            os.mkdir(tmpdir)
        execute(f'mv {self.args.name}.* {tmpdir}')
        with open("sim.done", 'w') as done:
            done.write('')



    def update_wrapper(self, new_reads: dict[str, str]) -> None:
        """
        Shared aspects of updates of aeons experiment
        :return:
        """
        # filter sequences with repeats at the end
        if self.args.filter_repeats:
            reads_filtered = self.repeat_filter.filter_batch(seq_dict=new_reads)
        else:
            reads_filtered = new_reads

        # load new sequences, incl length filter
        sequences = SequencePool(sequences=reads_filtered, min_len=self.args.min_seq_len)
        # add new sequences to AVA
        self.add_new_sequences(sequences=sequences)
        # check for overlaps and containment in pool
        self.overlap_pool()
        # trim sequences that might lead to overlaps
        self.trim_sequences()
        # call wrapper to update assembly
        contigs = self.assemble_add_and_filter_contigs()
        contig_pool = ContigPool(sequences=contigs.sequences)
        # write the current pool to file for mapping against
        self.pool.write_seq_dict(seq_dict=contigs.seqdict(), file=self.pool.contig_fa)
        # check if we have any frozen sequences
        frozen_ids = self.pool.decrease_temperature(lim=self.args.min_contig_len)
        self.remove_seqs(sequences=frozen_ids)
        # update the sequencing masks
        self.strat = contig_pool.process_contigs(
            score_vec=self.score_vec,
            ccl=self.rl_dist.approx_ccl,
            out_dir=self.out_dir,
            lam=self.rl_dist.lam,
            batch=self.batch)



    def process_batch_aeons(self, new_reads: dict[str, str], **kwargs) -> None:
        """
        pass-through function for superclass method
        :param new_reads: Dictionary of new sequence data
        :param kwargs: Catches other arguments
        :return:
        """
        self.update_wrapper(new_reads=new_reads)






import logging
from collections import defaultdict
from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from boss.core import Boss
from boss.mapper import Mapper
from boss.utils import adjust_length

from boss.runs.sequences import CoverageConverter, Scoring
from boss.runs.reference import Reference
from boss.runs.abundance_tracker import AbundanceTracker
from boss.runs.readstartdist import ReadStartDist




class BossRuns(Boss):


    def init(self) -> None:
        """
        Initialise necessary components for a BOSS-RUNS experiment
        These components are shared across the live and simulation experiments,
        i.e. this init function is also run in the init of the simulation

        :return:
        """
        # initialise reference
        self.ref = Reference(ref=self.args.ref, mmi=self.args.mmi, reject_refs=self.args.reject_refs)
        self.contigs = self.ref.contigs
        self.contigs_filt = {n: c for n, c in self.contigs.items() if not c.rej}
        # initialise a mapper using the reference
        self.mapper = Mapper(ref=self.ref.mmi)
        # initialise a translator for coverage conversions
        self.cc = CoverageConverter()
        # initialise the abundance tracker
        self.tracker = AbundanceTracker(contigs=self.contigs)
        # initialise the tracker for read start position distribution
        self.read_starts = ReadStartDist(contigs=self.contigs_filt)
        # initialise scoring array
        self.scoring = Scoring(ploidy=self.args.ploidy)
        self.scoring.init_score_array()
        # write initial strategies to file
        strat_dict = self.ref.get_strategy_dict()
        self._write_contig_strategies(contig_strats=strat_dict)



    def _write_contig_strategies(self, contig_strats: dict[str, NDArray]) -> None:
        """
        Write the strategies for all contigs to a single file.

        :param contig_strats: A dictionary containing the strategies for contigs.
        """
        cpath_tmp = f'{self.out_dir}/masks/boss_tmp.npz'
        np.savez(cpath_tmp, **contig_strats)
        # after writing to tmpfile, rename to replace the current strat
        cpath = f'{self.out_dir}/masks/boss.npz'
        Path(cpath_tmp).rename(cpath)
        # Example how to load these:
        # container = np.load(f'{cpath}.npz')
        # data = {key: container[key] for key in container}



    def _effect_increments(self, increments: defaultdict) -> None:
        """
        Loop through the increments of each contig and add the new coverage
        :param increments: defaultdict of lists of coverage counts per contig
        :return:
        """
        for cname, cont in self.contigs_filt.items():
            # grab the list of increments for this contig
            inc_list = increments[cname]
            cont.increment_coverage(increment_list=inc_list)



    def _update_scores_contigs(self) -> None:
        """
        Update the scores for each contig
        :return:
        """
        for cname, cont in self.contigs_filt.items():
            # main score updating function
            cont.scores, cont.entropy = self.scoring.update_scores(contig=cont)
            # modify scores after updating
            cont.modify_scores()


    def _check_buckets_contigs(self) -> bool:
        """
        Check if the buckets within the contigs have enough coverage to be switched on
        :return: Boolean if any strategy has been switched on
        """
        for cname, cont in self.contigs_filt.items():
            cont.check_buckets(threshold=self.args.bucket_threshold)
        # check if any switches are on
        switched_on = [c.switched_on for c in self.contigs.values()]
        return any(switched_on)


    def _update_benefits(self) -> None:
        """
        Update the expected benefit for all contigs
        :return:
        """
        for cname, cont in self.contigs_filt.items():
            cont.calc_smu()
            cont.calc_u(approx_ccl=self.rl_dist.approx_ccl)



    def _distribute_strategy(self, strat: NDArray, window: int = 100) -> None:
        """
        Place new decision strategies into the contigs strategies

        :param strat: Merged array of the updated strategy
        :param window: Downsampling window size
        :return:
        """
        i = 0
        for cname, cont in self.contigs_filt.items():
            # get the buckets of this contig and expand
            expand_fac = cont.bucket_size // window
            buckets_exp = np.repeat(cont.bucket_switches, expand_fac, axis=0)
            buckets = adjust_length(original_size=cont.strat.shape[0], expanded=buckets_exp)
            assert buckets.shape[0] == cont.strat.shape[0]
            # grab the new strategy
            cstrat = strat[i: i + cont.length // window, :]
            assert cstrat.shape == cont.strat.shape
            # assign new strat
            cont.strat[buckets, :] = cstrat[buckets, :]
            # log number of accepted sites
            f_perc = np.count_nonzero(cont.strat[:, 0]) / cont.strat.shape[0]
            r_perc = np.count_nonzero(cont.strat[:, 1]) / cont.strat.shape[0]
            logging.info(f'{cname}: {f_perc}, {r_perc}')
            i += cont.length // window


    def cleanup(self) -> None:
        """
        To be implemented

        :return:
        """
        pass


    def update_wrapper(self) -> None:
        """
        Second part of updates after a new batch of data.
        This is run after counting the coverage from new data

        :return:
        """
        # update the scores of all contigs
        self._update_scores_contigs()
        # flip strategy switches if threshold is reached
        switched_on = self._check_buckets_contigs()
        # UPDATE STRATEGY
        if switched_on:
            # update Fhat: unpack and normalise
            fhat_exp = self.read_starts.update_f_pointmass()
            self._update_benefits()
            # merge the benefits into one array for combined calculation
            benefit, smu = self.scoring.merge_benefit(self.contigs_filt)
            target_size = self.ref.n_sites // 100
            benefit_adj = adjust_length(original_size=target_size,
                                        expanded=benefit)
            smu_adj = adjust_length(original_size=target_size,
                                        expanded=benefit)
            fhat_adj = adjust_length(original_size=target_size,
                                     expanded=fhat_exp)
            assert fhat_adj.shape == benefit_adj.shape == smu_adj.shape
            # find the current decision strategy
            strat, threshold = self.scoring.find_strat_thread(
                benefit=benefit_adj,
                smu=smu_adj,
                fhat=fhat_adj,
                time_cost=self.rl_dist.time_cost
            )
            # distribute the strategy to the contigs
            self._distribute_strategy(strat=strat)
            # write strategies to file
            strat_dict = self.ref.get_strategy_dict()
            self._write_contig_strategies(contig_strats=strat_dict)



    def process_batch_runs(self, new_reads: dict[str, str], new_quals: dict[str, str]) -> None:
        """
        Process a batch of new data for the BOSS-RUNS mode
        This function is to be passed into the process_batch of the superclass

        :param new_reads: Dictionary of new sequences
        :param new_quals: Dictionary of qualities of new data
        :return:
        """
        # map the new reads to the reference
        paf_dict = self.mapper.map_sequences(sequences=new_reads)
        # convert coverage counts to increment arrays
        increments = self.cc.convert_records(paf_dict=paf_dict, seqs=new_reads, quals=new_quals)
        # effect the coverage increments for each contig
        self._effect_increments(increments=increments)
        # update the abundance tracker with new data
        self.tracker.update(n=len(new_reads), paf_dict=paf_dict)
        # update read starting position distribution
        self.read_starts.count_read_starts(paf_dict=paf_dict)
        # note: read length dist is updated in _get_new_data() of superclass
        self.update_wrapper()




import logging
from collections import defaultdict
from pathlib import Path
from string import ascii_letters

import numpy as np
from numpy.typing import NDArray
import bottleneck as bn
import mappy

from boss.utils import window_sum, adjust_length
from boss.mapper import Indexer
from boss.runs.sequences import Scoring




class Contig:

    def __init__(self, name: str, seq: str, ploidy: int = 1, rej: bool = False):
        """
        Initialise a contig object

        :param name: Name as in fasta header
        :param seq: DNA string of the contig.
        :param ploidy: Contig from haploid or diploid organism
        :param rej: Always reject from this contig?
        """
        self.name = name.strip().split(" ")[0]
        self.seq = seq.upper()
        self.length = len(self.seq)
        self.rej = rej   # flag whether to reject all reads from this contig
        self.seq_int = self._seq2int()
        self._init_coverage()
        self._init_buckets()
        self.scoring = Scoring(ploidy=ploidy)
        self.len_b = self.scoring.priors.len_b
        self.score0, self.ent0 = self.scoring.score0, self.scoring.ent0
        self._init_scores()
        self._init_strat()


    def _seq2int(self) -> NDArray:
        """
        Transform nucleotide strings to integer arrays.
        Unknown bases are set to integer 0, i.e. non-canonical nucleotides are not accounted for
        and get assigned an incorrect prior. This just means that more data might have to be
        collected to resolve the site. The alternative would be to implement priors for all IUPAC
        codes, which would be lots of work for very little gain.

        :return: Array of integers representing nucleotides
        """
        # define the dict for translating the bases
        transDict = defaultdict(str)
        transDict.update({'A': '0', 'C': '1', 'G': '2', 'T': '3'})
        for letter in ascii_letters:
            if letter not in transDict.keys():
                transDict[letter] = '0'
        # create translation table from dictionary
        base2int = str.maketrans(dict(transDict))
        # translate and convert from byte string
        read_integer = self.seq.translate(base2int)
        seq_int = np.frombuffer(read_integer.encode(), 'u1') - ord('0')
        assert len(set(seq_int[:100]) | {0, 1, 2, 3}) == 4
        return seq_int


    def _init_coverage(self) -> None:
        """
        Initialise containers to hold the observed coverage values

        :return:
        """
        self.coverage = np.zeros(shape=(self.length, 5), dtype="uint16")
        # to indicate where changes happened for score calculations
        self.change_mask = np.zeros(shape=self.length, dtype="bool")



    def _init_buckets(self, bucket_size: int = 20_000) -> None:
        """
        Each contig has multiple 'buckets' in which the strategy activates independently

        :param bucket_size: Size of the bucket windows
        :return:
        """
        self.bucket_size = bucket_size
        self.bucket_switches = np.zeros(shape=int(self.length // bucket_size) + 1, dtype="bool")
        self.switched_on = False



    def _init_scores(self) -> None:
        """
        Initialise the benefit scores at each site using the initial priors

        :return:
        """
        # initialise entropy and scores
        self.initial_scores = np.full(fill_value=self.score0[0], shape=self.length)
        self.scores = np.full(fill_value=self.score0[0], shape=self.length)
        self.entropy = np.full(fill_value=self.ent0[0], shape=self.length)



    def _init_strat(self, window: int = 100) -> None:
        """
        Initialise dummy strategies for the contig
        :param window: Downsampling window size
        :return:
        """
        if self.rej:
            self.strat = np.zeros(dtype="bool", shape=1)
        else:
            self.strat = np.ones(dtype="bool", shape=(self.length // window, 2))



    def increment_coverage(self, increment_list: list[int, int, NDArray, NDArray]) -> None:
        """
        Increment the coverage counts after parsing mappings

        :param increment_list: List of increment instructions for this contig
        :return:
        """
        # reset the change mask
        self.change_mask.fill(0)
        # temporary container for coverage
        tmp_cov = np.zeros(shape=self.coverage.shape, dtype="uint16")
        for (start, end, query_arr, addition) in increment_list:
            # range to index into the bit we want to increment
            indices = np.arange(query_arr.shape[0])
            # add the "addition" to the corresponding bases
            np.add.at(tmp_cov[start: end], (indices, query_arr), addition)
        # set the change mask
        self.change_mask[np.where(tmp_cov)[0]] = 1
        # add the coverage to the previous one
        self.coverage += tmp_cov



    def modify_scores(self) -> None:
        """
        Modify benefit scores after updating, specifically to ignore large deletions
        or missing regions in the sampled genome, e.g. accessory genes
        For each chromosome separately, in case there is abundance bias

        :return:
        """
        # coverage depth
        covsum = np.sum(self.coverage, axis=1)
        if np.mean(covsum) > 5:
            dropout_idx = self._find_dropout(covsum)
            logging.info(f'detected {dropout_idx.shape[0]} dropouts')
            self.scores[dropout_idx] = 0



    @staticmethod
    def _find_dropout(covsum: NDArray, mod: int = 8) -> NDArray:
        """
        If there are sites that have not had any coverage after some time
        we don't expect them to gain any at all and ignore their benefit

        :param covsum: coverage depth at each position
        :param mod: Threshold modifier
        :return: Array of dropout indices
        """
        cov_mean = np.mean(covsum)
        # ignore threshold is dependent on mean coverage
        threshold = int(cov_mean / mod)
        dropout = np.where(covsum <= threshold)[0]
        return dropout



    def check_buckets(self, threshold: float = 5.0) -> None:
        """
        Check if strategy needs to be activated in some of the buckets of a contig
        This happens if the mean coverage in a bucket is greater than the threshold

        :param threshold: when to switch on the strategy buckets
        :return:
        """
        # coverage depth
        csum = np.sum(self.coverage, axis=1)
        # coverage in buckets
        csum_buckets = window_sum(csum, self.bucket_size)
        cmean_buckets = np.divide(csum_buckets, self.bucket_size)
        cmean_buckets = adjust_length(original_size=self.bucket_switches.shape[0], expanded=cmean_buckets)
        # flip strategy switches
        self.bucket_switches[np.where(cmean_buckets >= threshold)] = 1
        switch_count = np.bincount(self.bucket_switches)
        states = len(switch_count)
        # log the first time a contig's strategy is switched on
        if states == 2 and not self.switched_on:
            self.switched_on = True
            logging.info(f"Activated strategy for: {self.name}")



    def calc_smu(self, window: int = 100, mu: int = 400) -> None:
        """
        Calculate S_mu, i.e. scores of mu-sized fragment at each position

        :param window: Size of downsampling window
        :param mu: Length of mu in model
        :return:
        """
        # downsample the scores
        self.scores_ds = np.zeros(shape=int(self.length // window) + 1)
        site_indices = np.arange(0, self.length) // window
        # avoid buffering
        np.add.at(self.scores_ds, site_indices, self.scores)
        # calculate smu - fwd needs double reversal due to how bn.move_sum() operates
        smu_fwd = bn.move_sum(self.scores_ds[::-1], window=mu // window, min_count=1)[::-1]
        smu_rev = bn.move_sum(self.scores_ds, window=mu // window, min_count=1)
        # assign smu as attribute
        self.smu = np.zeros(shape=(self.scores_ds.shape[0], 2))
        self.smu[:, 0] = smu_fwd
        self.smu[:, 1] = smu_rev



    def calc_u(self, approx_ccl: NDArray, window: int = 100) -> None:
        """
        Calculate the expected benefit of new fragments at each position
        New implementation using bn.move_sum

        :param approx_ccl: Approximation of read length distribution
        :param window: Downsampling window size
        :return:
        """
        # downsample read length dist
        approx_ccl_ds = approx_ccl // window
        mult = np.arange(0.05, 1, 0.1)[::-1]
        # temporary container
        tmp_benefit = np.zeros(shape=(self.scores_ds.shape[0], 2))
        for i in range(10):
            b_part_fwd = bn.move_sum(self.scores_ds[::-1], window=int(approx_ccl_ds[i]), min_count=1)[::-1]
            b_part_rev = bn.move_sum(self.scores_ds, window=int(approx_ccl_ds[i]), min_count=1)
            # apply weighting by length
            wgt = mult[i]
            tmp_benefit[:, 0] += (b_part_fwd * wgt)
            tmp_benefit[:, 1] += (b_part_rev * wgt)
        # assign as attribute
        self.expected_benefit = tmp_benefit
        self.additional_benefit = self.expected_benefit - self.smu
        # correct numerical issues
        self.additional_benefit[self.additional_benefit < 0] = 0




class Reference:

    def __init__(self, ref: str, mmi: str = None, reject_refs: str = None):
        """
        Initialise a reference object. Loads contigs, load or create index
        and set contigs from which to always reject

        :param ref: Path to the reference fasta file
        :param mmi: Optional path to minimap index file
        :param reject_refs: Optional comma-sep list of headers in fasta
        """
        self.ref = ref
        self.mmi = mmi
        if not Path(ref).is_file():
            raise FileNotFoundError("Reference file not found")
        ref_suff = Path(ref).suffixes
        if not any([r in {".fa", ".fasta"} for r in ref_suff]):
            raise ValueError("Reference needs to be either .fa or .fasta (optionally gzipped).")
        # create index if not present
        if not self.mmi:
            logging.info(f"Indexing reference: {self.ref}")
            mmi = f"{self.ref}.mmi"
            Indexer(fasta=self.ref, mmi=mmi)
            self.mmi = mmi
        else:
            if not Path(self.mmi).is_file():
                raise FileNotFoundError("Given mmi file not found")


        # headers of reference sequences from which to always reject
        if reject_refs:
            self.reject_refs = set(reject_refs.split(','))
        else:
            self.reject_refs = set()

        # load the sequences and their lengths
        logging.info("Reading reference file")
        self.contigs = self._load_contigs()
        self.n_sites = self._total_sites()



    def _load_contigs(self, min_len: int = 1e5, ploidy: int = 1) -> dict[str, Contig]:
        """
        Load contigs from fasta file

        :param min_len: Contigs shorter than this are ignored
        :param ploidy: Indicate haploid or diploid organism
        :return: dictionary of Contig objects
        """
        contigs = {}
        for cname, cseq, _ in mappy.fastx_read(self.ref):
            # filter short reference sequences that can cause issues
            if len(cseq) < min_len:
                continue
            # load reference sequences
            if cname not in self.reject_refs:
                contigs[cname] = Contig(name=cname, seq=cseq, ploidy=ploidy)
            # for ref seqs that we always reject, set sequence empty
            else:
                contigs[cname] = Contig(name=cname, seq="ACGT", ploidy=ploidy, rej=True)
        return contigs




    def _total_sites(self) -> int:
        """
        :return: Number of total sites in all contigs
        """
        return np.sum(list(self.contig_lengths().values()))


    def contig_sequences(self) -> dict[str, str]:
        """
        :return: Dictionary of contig sequences
        """
        return {c.name: c.seq for c in self.contigs.values()}


    def contig_lengths(self) -> dict[str, int]:
        """
        :return: Dictionary of contig names and their lengths
        """
        return {c.name: c.length for c in self.contigs.values()}



    def get_strategy_dict(self) -> dict[str, NDArray]:
        """
        Compile a dictionary of strategies from all contigs
        :return:
        """
        strat_dict = {}
        for cname, cont in self.contigs.items():
            strat_dict[cname] = cont.strat
        return strat_dict









from collections import defaultdict

import numpy as np
from numpy.typing import NDArray
from scipy.special import betaln

from boss.paf import Paf



class ReadStartDist:

    def __init__(self, contigs: dict, window_size: int = 2000, alpha: float = 1.0, p0: float = 0.1):
        """
        Initialise the probability distribution of fragment start sites

        :param contigs: Dictionary of contig objects
        :param window_size: Window size of counting read starts
        :param alpha: prior for alpha (hyperparameter for dirichlet prior)
        :param p0: Prior for sites with 0 observations
        """
        self.alpha = alpha
        self.p0 = p0
        self.window_size = window_size
        # track read start positions (forward and rev) in windows
        self.read_starts = {cname : np.zeros(shape=(int(c.length / window_size), 2)) for cname, c in contigs.items()}
        # fhat exists only in its merged form, i.e. for use in updating on a merged array
        self.total_len = np.sum([a.shape[0] for a in self.read_starts.values()])
        self.target_size = int(np.sum([c.length for c in contigs.values()]) // 100)
        self.on_target = 1   # TODO
        self.fhat = self.update_f_pointmass()



    def merge(self) -> NDArray:
        """
        :return: Concatenated array of read starting sites across all contigs
        """
        return np.concatenate(list(self.read_starts.values()))



    def count_read_starts(self, paf_dict: dict[str, list]) -> None:
        """
        Keep track of the read starting positions C_{i,o}, used to update F
        Read starts are saved in non-overlapping windows of windowsize

        :param paf_dict: Dictionary of read mappings
        :return:
        """
        # collect all starting positions
        starts_fwd = defaultdict(list)
        starts_rev = defaultdict(list)

        for rid in paf_dict.keys():
            rec = paf_dict[rid]
            # choose the highest ranked mapping
            if len(rec) > 1:
                rec = Paf.choose_best_mapper(rec)[0]
            else:
                rec = rec[0]

            if rec.rev:
                starts_rev[rec.tname].append(rec.tend)
            else:
                starts_fwd[rec.tname].append(rec.tstart)

        for cname, r_starts in self.read_starts.items():
            # count the number of read starts in windows
            n_windows = int(r_starts.shape[0])
            bins_fwd = np.histogram(
                starts_fwd[cname],
                bins=n_windows,
                range=(0, self.window_size * n_windows))[0].astype(dtype='float')
            bins_rev = np.histogram(
                starts_rev[cname],
                bins=n_windows,
                range=(0, self.window_size * n_windows))[0].astype(dtype='float')

            # add new counts to the array
            self.read_starts[cname][:, 0] += bins_fwd
            self.read_starts[cname][:, 1] += bins_rev



    def update_f_pointmass(self) -> NDArray:
        """
        Bayesian approach to update posterior values of f_hat using counts of read starting positions.
        Version with point mass at 0 (for sites with C == 0)

        :return: Array of read starting probabilities for all positions
        """
        # concatenate arrays
        merged = self.merge()
        n_windows = merged.shape[0]
        fhat = np.zeros(shape=merged.shape)
        # First, sites with C > 0
        nonzero_indices = np.nonzero(merged)
        nonzero = merged[nonzero_indices]
        num = np.add(self.alpha, nonzero)
        Csum = np.sum(nonzero)
        denom = 2 * n_windows * self.alpha + Csum
        fhat[nonzero_indices] = np.divide(num, denom)
        # then sites with C == 0
        rhs = (self.alpha / (2 * n_windows * self.alpha + Csum))
        beta_num = np.exp(betaln(self.alpha, ((2 * n_windows - 1) * self.alpha + Csum)))
        beta_denom = np.exp(betaln(self.alpha, ((2 * n_windows - 1) * self.alpha))) or 1e-20
        p0_bit = self.p0 / (self.p0 + (1 - self.p0))
        lhs = 1 - p0_bit * (beta_num / beta_denom)
        expectedPost = lhs * rhs
        # mask for the zero count sites - derived from nonzero indices
        zero_indices = np.ones(shape=fhat.shape, dtype="bool")
        zero_indices[nonzero_indices] = 0
        fhat[zero_indices] = expectedPost
        # expand from downsampled size
        fhat_exp = self._expand_fhat(fhat)
        return fhat_exp



    def _expand_fhat(self, fhat, downsample_window: int = 100) -> NDArray:
        """
        Expand and normalise Fhat from the compacted length of genome_length / windowSize

        :param fhat: probabilities of read starting positions
        :param downsample_window: length of windows used in generating decision strategies
        :return: read start probs expanded from genome_length / windowSize
        """
        nrep = int(self.window_size // downsample_window)
        # expand to the downsampled size
        fhat_exp = np.repeat(fhat, nrep, axis=0)
        # correct for small length difference
        lendiff = self.target_size - fhat_exp.shape[0]
        assert lendiff < self.window_size
        if lendiff > 0:
            # genome is longer than Fhat
            fhat_exp = np.append(fhat_exp, fhat_exp[-lendiff:], axis=0)
        elif lendiff < 0:
            # genome is shorter than Fhat
            fhat_exp = fhat_exp[:-abs(lendiff)]
        else:
            pass

        # normalise only if not empty
        fhat_sum = np.sum(fhat_exp)
        if fhat_sum != 0:
            # normalise for ratio of on/off target reads
            # uses estimated on-target proportion from initially accepted sites
            normalizer = self.on_target / fhat_sum     # TODO get the on-target estimator
            fhat_exp = np.multiply(fhat_exp, normalizer)
        return fhat_exp



    def estimate_priors(self) -> tuple[float, float]:
        """
        Estimate alpha, the concentration hyperparameter of the dirichlet prior,
        by equating the variance of Fhat with the variance of the dirichlet.

        :return: Prior of dirichlet and proportion of gap sites
        """
        merged = self.merge()
        n_windows = merged.shape[0]
        # filter readStartCounts for positions with 0
        zeroCounts = np.count_nonzero(merged == 0)
        p0 = zeroCounts / (n_windows * 2)
        # get sum of mapped reads
        Csum = np.sum(merged) or 1e-30
        # simple estimator of alpha
        Fhat = np.divide(merged, Csum)
        # variance of Fhat to equate with var of D(alpha)
        Vhat = np.var(Fhat, ddof=0) or 1e-30
        # if Vhat is very small, alpha becomes very large
        lhs = (2 * n_windows - 1) / (Vhat * 8 * (n_windows ** 3))
        rhs = 1 / (2 * n_windows)
        alpha = lhs - rhs
        return alpha, p0






import logging

import numpy as np

from boss.mapper import Mapper
from boss.utils import find_blocks_ge
from boss.aeons.sequences import SequencePool




class RepeatFilter:
    """
    A class to:
     - extract repeats
     - generate a fasta
     - map batches to repeat library for filtering

    """

    def __init__(self, name: str, seqpool: SequencePool):
        """
        Initialise a RepeatFilter, incl indexing the sequences, chopping and mapping them

        :param name: Name of experiment
        :param seqpool: Sequence pool of initial data
        """
        self.seqpool = seqpool
        self.name = name
        self.library = f'{name}.repeat_lib.fa'
        # initialise a mapper against the long reads
        SequencePool.write_seq_dict(seqpool.seqdict(), f'{name}.seqs.fa')
        lr_mapper = Mapper(ref=f'{name}.seqs.fa')
        # chop and map all sequences
        little_seqs = self._chop_seqs()
        mappings = lr_mapper._mappy_batch(sequences=little_seqs)
        covs = self._count_cov(mappings)
        self.covs = covs
        # # DEBUG write mappings to file
        # with open("little_seqs_mappings.paf", 'w') as lsm:
        #     lsm.write(mappings)
        # import pickle
        # with open("coverage.pkl", 'wb') as covpkl:
        #     pickle.dump(covs, covpkl)
        # find the min cov for repeats
        self._find_limit()
        # find repeats
        repeat_blocks = self._identify_repeat_sites()
        # write to library file
        self._write_repeat_seqs(repeat_blocks)


    def _chop_seqs(self, window: int = 100, step: int = 100) -> dict:
        """
        Chop the current sequences into smaller bits using a sliding window

        :param window: Size of sliding window
        :param step: Stepsize of sliding windows
        :return: Dictionary of chopped sequences to map
        """
        seqs = self.seqpool.seqdict()
        little_seqs = {}
        for header, seq in seqs.items():
            i = 0
            while i < len(seq):
                little_seqs[f'{header}-{i:010}'] = seq[i: i + window]
                i += step
        with open('little_seqs.fa', 'w') as fh:
            for header, seq in little_seqs.items():
                fh.write(f'>{header}\n')
                fh.write(f'{seq}\n')
        return little_seqs


    @staticmethod
    def _count_cov(mappings: str) -> dict:
        """
        loop through mappings and count the coverage of all targets

        :param mappings: String of a PAF file
        :return: Dict of coverages
        """
        covs = {}
        for line in mappings.split('\n'):
            rec = line.split('\t')
            # check if target array exists
            if not rec[5] in covs.keys():
                covs[rec[5]] = np.zeros(shape=int(rec[6]))
            # grab the cov
            c = covs[rec[5]]
            c[int(rec[7]): int(rec[8])] += 1
        return covs



    def _find_limit(self) -> None:
        """
        Find the limit of coverage to consider repeat sequence
        :return:
        """
        # find the max value first
        maximum = 0
        for c in self.covs.values():
            cmax = np.max(c)
            if cmax > maximum:
                maximum = cmax

        # count all coverage values
        bcounts = np.zeros(int(np.max(maximum) + 1), dtype="int")
        for c in self.covs.values():
            c[0] = 0  # make sure count starts at 0
            bcounts_arr = np.bincount(c.astype('int'))
            for i in range(len(bcounts_arr)):
                bcounts[i] += bcounts_arr[i]

        # limit
        lim = np.quantile(np.repeat(np.arange(len(bcounts)), repeats=bcounts), 0.999)
        if lim < 3:
            lim = 3.0
        self.lim = lim


    def _identify_repeat_sites(self) -> dict:
        """
        find positions where supposed repeats are in sequences
        :return: Dict of repeat blocks for each header
        """

        repeat_blocks = {}
        for header, cov in self.covs.items():
            blocks = find_blocks_ge(cov, self.lim, min_len=100)
            if len(blocks) > 0:
                repeat_blocks[header] = blocks
        return repeat_blocks


    def _write_repeat_seqs(self, repeat_blocks: dict) -> None:
        """
        Write the repeat seqs to file and collect them in dictionary

        :param repeat_blocks: Dict of repeat block coordinates
        :return:
        """
        n_seqs = 0
        repeats = {}
        with open(self.library, 'w') as fh:
            for header, blocks in repeat_blocks.items():
                for start, end in blocks:
                    r = Repeat(header, start, end)
                    r.get_sequence(seqpool=self.seqpool.sequences)
                    fa = r.fasta()
                    fh.write(fa)
                    repeats[r.header] = r.seq
                    n_seqs += 1
        self.repeats = repeats


    @staticmethod
    def _check_coverage(rep_cov: dict, window: int = 500) -> set:
        """
        Check whether a read has a potential repeat on either end

        :param rep_cov: Dict of coverage counts
        :param window: size of end windows
        :return: set of read ids with potential repeats at end
        """
        danger = set()
        for header, rcov in rep_cov.items():
            beginning = rcov[: window]
            if np.sum(beginning) > 5:
                danger.add(header)
            ending = rcov[window:]
            if np.sum(ending) > 5:
                danger.add(header)
        return danger



    def filter_batch(self, seq_dict: dict) -> dict:
        """
        Check a dict of input sequences against a repeat library

        :param seq_dict: Dict of input sequences
        :return: Dict of sequences with potential repeat-seqs removed
        """
        logging.info("repeat filtering batch of reads")
        # write batch to file
        bfile = f'{self.name}.batch.fa'
        with open(bfile, 'w') as fh:
            for header, seq in seq_dict.items():
                fa = f'>{header}\n{seq}\n'
                fh.write(fa)
        # initialise a LinearMapper object
        lm = Mapper(ref=bfile)
        # first map them to the library
        mappings = lm._mappy_batch(self.repeats)
        rep_cov = self._count_cov(mappings)
        danger_ids = self._check_coverage(rep_cov)
        filtered_seqs = {h: s for h, s in seq_dict.items() if h not in danger_ids}
        return filtered_seqs




class Repeat:

    def __init__(self, rid: str = None, start: int = 0, end: int = -1):
        """
        Initialise a repeat object

        :param rid: ID of source sequence
        :param start: Start pos on source sequence
        :param end: End pos on source sequence
        """
        self.rid = rid
        self.start = start
        self.end = end
        self.seq = ''


    def get_sequence(self, seqpool: dict):
        """
        Get sequence from Sequencepool dict using ID and coordinates

        :param seqpool: Dict of SequencePool, i.e. SequencePool.sequences
        :return:
        """
        # index into seqpool and trim
        try:
            self.seq = seqpool[self.rid].seq[self.start: self.end]
        except KeyError:
            logging.info(f'{self.rid} not found in seqpool')
            return


    def fasta(self) -> str:
        """
        Generate fasta representation of itself

        :return: string representation in FASTA format
        """
        if not self.seq:
            return ""
        # construct fasta entry
        self.header = f'{self.rid}-{self.start}:{self.end}'
        fa = f'>{self.header}\n{self.seq}\n'
        return fa









import time
import logging
from pathlib import Path
from types import SimpleNamespace

from boss.live import LiveRun, Sequencer
from boss.readlengthdist import ReadlengthDist
from boss.batch import FastqBatch




class Boss:

    def __init__(self, args: SimpleNamespace):
        """
        Main initialisation of experiment

        :param args: Configuration of the experiment
        """
        self.args = args
        self.name = args.name
        self.processed_files = set()
        self.n_fastq = 0
        self.batch = 0
        # initial strategy is to accept
        self.strat = 1
        # initialise output directory structure
        self._init_file_struct()
        # object to record read lengths
        self.rl_dist = ReadlengthDist()



    def _init_file_struct(self) -> None:
        """
        Set up the required directory structure

        :return:
        """
        # make sure the run name does not have any spaces
        assert ' ' not in self.args.name

        self.out_dir = f'./out_{self.args.name}'
        out_path = Path(self.out_dir)
        out_path.mkdir(parents=True, exist_ok=True)
        (out_path / "masks").mkdir(parents=True, exist_ok=True)
        (out_path / "fq").mkdir(parents=True, exist_ok=True)
        (out_path / "logs").mkdir(parents=True, exist_ok=True)
        (out_path / "contigs").mkdir(parents=True, exist_ok=True)
        (out_path / "contigs" / "prev").mkdir(parents=True, exist_ok=True)
        (out_path / "contigs" / "init").mkdir(parents=True, exist_ok=True)
        (out_path / "metrics").mkdir(parents=True, exist_ok=True)


    def _init_live(self) -> None:
        """
        Wrapper to initialise a live experiment, incl. device related things
        find output path where the fastq files are placed
        find channels toml for runs with multiple regions

        :return:
        """
        # connect to sequencing machine
        # early exit for pytests
        if self.args.device == "TEST":
            sequencer = Sequencer()

        else:
            sequencer = LiveRun.connect_sequencer(device=self.args.device,
                                                  host=self.args.host,
                                                  port=self.args.port)
            sequencer.grab_channels(run_name=self.args.name)

        # get the relevant infos from the Sequencer
        self.args.fq = f'{sequencer.out_path}/fastq_pass'
        assert Path(self.args.fq).is_dir()
        # readfish needs to have placed the channels.toml at this point
        # channels can be an empty set if there is only one condition
        # then data from all channels will be used (i.e. later regex skipped)
        self.channels = sequencer.channels



    def _get_new_data(self) -> tuple[dict, dict]:
        """
        Get new read data and parse it into a batch object

        :return: Dictionaries for read sequences and their qualities
        """
        # find new fastq files
        new_fastq = LiveRun.scan_dir(
            fastq_pass=self.args.fq, processed_files=self.processed_files)
        if not new_fastq:
            logging.info("no new files, deferring update ")
            return {}, {}
        # add the new files to the set of processed files
        self.processed_files.update(new_fastq)
        self.n_fastq += len(new_fastq)
        fq_batch = FastqBatch(fq_files=new_fastq, channels=self.channels)
        # read length dist is updated straight away, not in func of subclass
        self.rl_dist.update(read_lengths=fq_batch.read_lengths)
        return fq_batch.read_sequences, fq_batch.read_qualities


    def _time_to_next_update(self, tic: float) -> int:
        """
        Calculate how long to wait until the next update given duration of previous

        :param tic: Start time of previous update
        :return: Seconds to wait until next update
        """
        toc = time.time()
        passed = toc - tic
        next_update = int(self.args.wait - passed)
        logging.info(f"batch took: {passed}")
        logging.info(f"finished update, waiting for {next_update}s ... \n")
        return next_update


    def launch_live_components(self):
        # launch readfish and initialise connection to sequencer if live experiment
        if hasattr(self.args, "live_run") and getattr(self.args, "live_run"):
            LiveRun.launch_readfish(
                toml=self.args.toml_readfish,
                device=self.args.device,
                name=self.name
            )
            self._init_live()


    def process_batch(self, main_processing_func: callable) -> int:
        """
        Main processing function for all live BOSS experiments.
        Variants can pass processing functions in

        :param main_processing_func:
        :return: Seconds to wait until next update
        """
        logging.info("")
        logging.info(f"Next batch ---------------------------- # {self.batch}")
        tic = time.time()

        new_reads, new_quals = self._get_new_data()
        if not new_reads:
            return self.args.wait

        main_processing_func(new_reads=new_reads, new_quals=new_quals)

        next_update = self._time_to_next_update(tic)
        self.batch += 1
        return next_update


    def process_batch_sim(self, main_processing_func: callable) -> int:
        """
        Main entry point for simulated experiments. Variants can
        pass processing functions in. A sampler is also required.

        :param main_processing_func:
        :return: Seconds to wait until next update
        """
        logging.info("")
        logging.info(f"Next batch ---------------------------- # {self.batch}")
        tic = time.time()

        main_processing_func()

        next_update = self._time_to_next_update(tic)
        self.batch += 1
        return next_update











import subprocess
from types import SimpleNamespace
from importlib.metadata import version

import numpy as np
from numpy.typing import NDArray
import psutil


def empty_file(path: str) -> None:
    """
    Create an empty file at the specified path.

    :param path: The path to the file.
    """
    with open(path, 'w'):
        pass
    return


def init_logger(logfile: str, args: SimpleNamespace) -> None:
    """
    Initialize the logger with the given logfile and log the arguments.

    :param logfile: The path to the logfile.
    :param args: The arguments to log.
    """
    empty_file(logfile)
    import logging
    logging.basicConfig(format='%(asctime)s %(message)s',
                        level=logging.INFO,
                        handlers=[logging.FileHandler(f"{logfile}"), logging.StreamHandler()])

    logging.info("BOSS RUNS/AEONS")
    logging.info(f"{version('boss_runs')}")
    logging.info('\n')
    for a, aval in args.__dict__.items():
        logging.info(f'{a} {aval}')
    logging.info('\n')


def execute(command: str) -> tuple[str, str]:
    """
    Execute a command in a shell and return the stdout and stderr.

    :param command: The command to execute.
    :return: The stdout and stderr as a tuple.
    """
    # create the unix process
    running = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               encoding='utf-8', shell=True)
    # run on a shell and wait until it finishes
    stdout, stderr = running.communicate()
    return stdout, stderr



def write_logs(stdout: str, stderr: str, basename: str) -> None:
    """
    Write the stdout and stderr from a subprocess to files.

    :param stdout: The stdout output.
    :param stderr: The stderr output.
    :param basename: The base name for the log files.
    """
    # write stdout and stderr from subprocess to file
    with open(f'{basename}.out', 'a') as outf:
        outf.write(stdout)
        outf.write('\n')
    with open(f'{basename}.err', 'a') as errf:
        errf.write(stderr)
        errf.write('\n')


def binc(arguments: tuple[NDArray, NDArray]) -> NDArray:
    """
    Helper function for multithreaded bincounting

    :param arguments: Container of array and weights to count
    :return:
    """
    a, b = arguments
    res = np.bincount(a, weights=b)
    return res


def reverse_complement(dna: str) -> str:
    '''
    Return the reverse complement of a dna string. Used when parsing the cigar
    of a read that mapped on the reverse strand.

    :param dna: string of characters of the usual alphabet
    :return: reverse complemented input dna
    '''
    trans = str.maketrans('ATGC', 'TACG')
    rev_comp = dna.translate(trans)[::-1]
    return rev_comp


def random_id(k: int = 20) -> str:
    """
    Generate a random alphanumeric ID of length k.

    :param k: The length of the ID.
    :return: The generated random ID.
    """
    import random
    import string
    x = ''.join(random.choices(string.ascii_letters + string.digits, k=k))
    return x


def load_gfa(gfa_path: str) -> dict[str, str]:
    """
    Load a GFA file and return a dictionary of sequences.

    :param gfa_path: The path to the GFA file.
    :return: A dictionary mapping header strings to sequence strings.
    """
    def _load_gfa(infile: str):
        with open(infile, 'r') as gfa_file:
            for line in gfa_file:
                if line.startswith('S'):
                    ll = line.split('\t')
                    header = ll[1]
                    seq = ll[2]
                    yield header, seq

    sequences = {}
    for hd, sq in _load_gfa(gfa_path):
        sequences[hd] = sq
    return sequences


# def find_blocks_generic(arr: NDArray, x: int, min_len: int) -> NDArray:
#     """
#     Find blocks in the array that match x.
#
#     :param arr: The input array.
#     :param x: The value to find blocks of.
#     :param min_len: The minimum length of blocks to report.
#     :return: An array containing the start and end positions of the blocks.
#     """
#     # find run starts
#     x_pos = np.where(arr == x)[0]
#
#     if x_pos.shape[0] == 0:
#         return np.array([])
#
#     # diff between neighboring loc
#     x_diff = np.diff(x_pos)
#     # if diff > 1: new block
#     big_dist = np.where(x_diff > 1)[0]
#     # the first entry is a block start and then all other where a big change happens
#     # also each change is a block end, and the last entry of course as well
#     block_ranges = np.concatenate((np.array([x_pos[0]]), x_pos[big_dist + 1],
#                                    x_pos[big_dist] + 1, np.array([x_pos[-1] + 1])))
#     blocks = block_ranges.reshape(big_dist.shape[0] + 1, 2, order='F')
#     # only report blocks longer than min_len
#     blocks_filt = blocks[np.where(blocks[:, 1] - blocks[:, 0] > min_len)[0], :]
#     return blocks_filt


def find_blocks_ge(arr: NDArray, x: float, min_len: int) -> NDArray:
    """
    Find blocks in the array where the value is greater than or equal to x.

    :param arr: The input array.
    :param x: The value to compare against.
    :param min_len: The minimum length of blocks to report.
    :return: An array containing the start and end positions of the blocks.
    """
    # find run starts
    x_pos = np.where(arr >= x)[0]

    if x_pos.shape[0] == 0:
        return np.array([])

    # diff between neighboring loc
    x_diff = np.diff(x_pos)
    # if diff > 1: new block
    big_dist = np.where(x_diff > 1)[0]
    # the first entry is a block start and then all other where a big change happens
    # also each change is a block end, and the last entry of course as well
    block_ranges = np.concatenate((np.array([x_pos[0]]), x_pos[big_dist + 1],
                                   x_pos[big_dist] + 1, np.array([x_pos[-1] + 1])))
    blocks = block_ranges.reshape(big_dist.shape[0] + 1, 2, order='F')
    # only report blocks longer than min_len
    blocks_filt = blocks[np.where(blocks[:, 1] - blocks[:, 0] > min_len)[0], :]
    return blocks_filt



def window_sum(arr: NDArray, w: int) -> NDArray:
    """
    Calculate sums of non-overlapping windows

    :param arr: Input array
    :param w: Window size to calculate sums in
    :return: Reduced array of window sums
    """
    sumw = np.sum(arr[: (len(arr) // w) * w].reshape(-1, w), axis=1)
    return sumw



def adjust_length(original_size: int, expanded: NDArray) -> NDArray:
    """
    Adjust the size of an array after expanding during downsampling

    :param original_size: Target size of expanded array
    :param expanded: Input array after expansion
    :return: Adjusted input array
    """
    # after expanding arrays from binning
    lendiff = original_size - expanded.shape[0]
    if lendiff > 0:
        # original is longer than replacement
        repl = np.append(expanded, expanded[-lendiff:], axis=0)
    elif lendiff < 0:
        # original is shorter than replacement
        repl = expanded[: -abs(lendiff)]
    else:
        repl = expanded

    assert repl.shape[0] == original_size
    return repl




def search_running_process(processName: str) -> bool:
    '''
    Check if there is any running process that contains the given name processName.
    adapted from https://gist.github.com/Sanix-Darker/8cbed2ff6f8eb108ce2c8c51acd2aa5a

    :param processName: The name of the process to check.
    '''
    # Iterate over the all running processes
    for proc in psutil.process_iter():
        try:
            if processName.lower() in proc.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False






"""Main entry point for command line read until scripts.

Set as entrypoint in ``pyproject.toml``

TODO MODIFIED version of _cli_base.py from readfish to setup arguments
TODO returns args instead of launching entry-point, otherwise identical
"""
# from __future__ import annotations
import sys
import argparse
import importlib
import logging

from readfish.__about__ import __version__
from readfish._loggers import setup_logger, print_args


def main(argv: list[str] | None = None):   # -> None:
    """Main function for entry point of the read until scripts.

    :param argv: used in tests default None
    :raises SystemExit: Raises a system exit when the command function exits.
    """
    parser = argparse.ArgumentParser(
        prog="readfish",
        epilog="See '<command> --help' to read about a specific sub-command.",
        allow_abbrev=False,
    )
    version = f"readfish {__version__}"
    parser.add_argument("--version", action="version", version=version)
    subparsers = parser.add_subparsers(dest="command", help="Sub-commands", metavar="")
    # add new entry point here
    cmds = [
        ("targets", "targets"),
        ("barcode-targets", "targets"),
        ("unblock-all", "unblock_all"),
        ("validate", "validate"),
        ("stats", "stats"),
    ]
    # Entry point is imported during runtime, and added as a sub command to readfish
    for cmd, module in cmds:
        _module = importlib.import_module(f"readfish.entry_points.{module}")
        _parser = subparsers.add_parser(cmd, help=_module._help)
        for *flags, opts in _module._cli:
            _parser.add_argument(*flags, **opts)
        _parser.set_defaults(func=_module.run)

    args, extras = parser.parse_known_args(argv)

    if args.command is not None:
        logger = setup_logger(
            "readfish",
            level=getattr(logging, args.log_level.upper()),
            log_format=args.log_format,
            log_console=True,
            log_file=args.log_file,
            # FIXME: Need to evaluate whether we should be propagating this or not?
            #        pytest caplog doesn't work without it..., but will a chunk/other
            #        debug log create issues?
            # propagate=True,
        )
        logger.info(" ".join(sys.argv) if argv is None else argv)
        print_args(args, printer=logger.info, exclude=["func"])
        # TODO raise SystemExit(args.func(parser, args, extras))
        return parser, args
    else:
        parser.print_help()


if __name__ == "__main__":
    main()





from io import StringIO
from collections import defaultdict
from types import SimpleNamespace
from pathlib import Path

import numpy as np





class PafLine:
    '''
    @DynamicAttrs
    parse a single alignment from a PAF into a flexible container
    '''

    def __init__(self, line: str, tags: bool = True):
        """
        Parse a line from a PAF file

        :param line: string representation of a line in a PAF file
        :param tags: boolean indicator whether to parse tags
        """
        self.line = line
        fields = ['qname', 'qlen', 'qstart', 'qend',
                  'strand', 'tname', 'tlen', 'tstart', 'tend',
                  'num_matches', 'alignment_block_length',
                  'mapq']
        core = 12
        record = line.strip().split("\t")

        f = PafLine.format_records(record[:core])
        for i in range(core):
            setattr(self, fields[i], f[i])

        # make sure query and target name are strings
        self.qname = str(self.qname)
        self.tname = str(self.tname)

        self.rev = 0 if self.strand == '+' else 1
        # parse the tags only if needed
        if tags:
            tags_parsed = PafLine.parse_tags(record[core:])
            self.align_score = int(tags_parsed.get("AS", 0))
            self.cigar = tags_parsed.get("cg", None)
            self.s1 = tags_parsed.get("s1", 0)
            prim = tags_parsed.get("tp", None)
            self.primary = 1 if prim == 'P' else 0
            # self.dv = tags_parsed.get('dv', 0)

        # markers for trimming
        self.qprox = False
        self.tprox = False
        # dummy inits
        self.maplen = None
        self.min_length_pair = None


    @staticmethod
    def format_records(record: list) -> list:
        """
        Helper function to make fields of a PafLine the right type

        :param record: Split string of PAFline into list
        :return: Same list but with types converted to int
        """
        return [PafLine.conv_type(x, int) for x in record]


    @staticmethod
    def parse_tags(tags: list) -> dict:
        """
        Parse tags of a PAFline into a dictionary

        :param tags: List of SAM style tags
        :return: Dict of SAM style tags
        """
        c = {"i": int, "A": str, "f": float, "Z": str}
        return {
            key: PafLine.conv_type(val, c[tag])
            for key, tag, val in (x.split(":") for x in tags)
        }


    @staticmethod
    def conv_type(s: str, func: callable):
        """
        Generic converter, to change strings to other types

        :param s: Input string to convert to a different type
        :param func: Target type of input string
        :return: Either converted or original type
        """
        try:
            return func(s)
        except ValueError:
            return s


    def filter(self, filters: SimpleNamespace) -> bool:
        """
        Check if an alignment needs to be filtered out

        :param filters: Filters stored as arguments
        :return: Indicator whether line is filtered
        """
        # like classify, pack all conditions in here
        if self._self_aligned():
            return True
        if self.map_length() < filters.min_map_len:
            return True
        if self.s1 < filters.min_s1:
            return True
        if self.min_length_in_pair() < filters.min_seq_len:
            return True
        # if none of the filters triggered
        return False


    def min_length_in_pair(self) -> float:
        """
        Get the shorter length of the two aligned sequences

        :return: Length of shorter sequence in pair
        """
        if not self.min_length_pair:
            self.min_length_pair = min(self.qlen, self.tlen)
        return self.min_length_pair


    def overhang(self) -> float:
        """
        Get the sum of the smallest overhangs on both sequences
        Used in classification of mapping

        :return: Sequence overhang in alignment
        """
        if not self.rev:
            overhang = min(self.qstart, self.tstart) +\
                       min(self.qlen - self.qend, self.tlen - self.tend)
        else:
            overhang = min(self.qstart, self.tlen - self.tend) +\
                       min(self.tstart, self.qlen - self.qend)
        return overhang


    def map_length(self) -> float:
        """
        Get the shorter alignment length. I.e. either the span on query or target

        :return: Mapping length of the paf record
        """
        if not self.maplen:
            self.maplen = min(self.qend - self.qstart, self.tend - self.tstart)
        return self.maplen


    def classify(self) -> int:
        """
        Classify the alignment according to miniasm algorithm 5
        Records that are being classified have already been filtered

        :return: Indicator of alignment type
        """
        c = -1

        if self._internal_match():
            c = 1
        elif self._first_contained():
            c = 2
        elif self._second_contained():
            c = 3
        elif self._first_contained_fallback():
            c = 2
        elif self._second_contained_fallback():
            c = 3
        else:
            pass

        # if still unclassified -> overlap
        if c < 0:
            c, qside, tside = self._overlap_orientation()
            self.qside = qside
            self.tside = tside

        # second chance for internal matches
        # consider them as ovl under special circumstances
        if c == 1:
            if self._first_contained_mostly():
                c = 2
            elif self._second_contained_mostly():
                c = 3
            elif self.internal_match_is_overlap():
                c = 6  # class 6: needs trimming
            else:
                pass

        return c



    def _internal_match(self) -> bool:
        """
        Check if the alignment is an internal match.
        We consider a maximum of 15 percent of the map length as limit

        :return: indicator if internal match
        """
        if self.overhang() > (self.map_length() * 0.15):
            return True
        else:
            return False



    def _first_contained(self) -> bool:
        """
        First contained means the query is contained in the target.
        This can be different in either forward or reverse relative orientation

        :return: indicator for query contained in target
        """
        # FORWARD
        if not self.rev:
            if (self.qstart <= self.tstart) and ((self.qlen - self.qend) < (self.tlen - self.tend)):
                return True
            else:
                return False
        # REVERSE
        else:
            if (self.qstart <= (self.tlen - self.tend)) and ((self.qlen - self.qend) < self.tstart):
                return True
            else:
                return False



    def _second_contained(self) -> bool:
        """
        Second contained means the target is contained in the query.
        This can be different in either forward or reverse relative orientation

        :return: indicator for target contained in query
        """
        # FORWARD
        if not self.rev:
            if (self.qstart >= self.tstart) and ((self.qlen - self.qend) > (self.tlen - self.tend)):
                return True
            else:
                return False
        # REVERSE
        else:
            if (self.qstart >= (self.tlen - self.tend)) and ((self.qlen - self.qend) > self.tstart):
                return True
            else:
                return False



    def _first_contained_fallback(self) -> bool:
        """
        Handling edge cases of containment: if more than 90 percent of the sequence
        is covered my a mapping, then we assume contained

        :return: Indicator for query contained in target
        """
        qcov = self.qend - self.qstart
        if (qcov / self.qlen) >= 0.90:
            return True
        else:
            return False



    def _second_contained_fallback(self) -> bool:
        """
        Handling edge cases of containment: if more than 90 percent of the sequence
        is covered my a mapping, then we assume contained

        :return: Indicator for target contained in query
        """
        tcov = self.tend - self.tstart
        if (tcov / self.tlen) >= 0.90:
            return True
        else:
            return False



    def _first_contained_mostly(self) -> bool:
        """
        This is to consider internal matches of long sequences as containments instead.
        This is done so that we can count the coverage information of such events

        :return: indicator for mostly-contained internal matches
        """
        qcov = self.qend - self.qstart
        if (qcov / self.qlen) >= 0.50 and self.qlen > 20000:
            return True
        else:
            return False



    def _second_contained_mostly(self) -> bool:
        """
        This is to consider internal matches of long sequences as containments instead.
        This is done so that we can count the coverage information of such events

        :return: indicator for mostly-contained internal matches
        """
        tcov = self.tend - self.tstart
        if (tcov / self.tlen) >= 0.50 and self.qlen > 20000:
            return True
        else:
            return False



    def _overlap_orientation(self) -> tuple[int, str, str]:
        """
        Identify details of overlaps, i.e. which overlaps which and on which end

        :return: Indicator for overlap source and their respective ends
        """
        if not self.rev:
            if self.qstart > self.tstart:
                # 'A' overlaps 'B'
                # a + b +
                return 4, 'R', 'L'
            else:
                # B overlaps A
                # b + a +
                return 5, 'L', 'R'
        elif self.qstart > (self.qlen - self.qend):
            if self.qstart > (self.tlen - self.tend):
                # 'A' overlaps 'B'
                # a + b -
                return 4, 'R', 'R'    # should this be LR?
            else:
                # B overlaps A
                # b + a -
                return 5, 'R', 'R'
        elif (self.qlen - self.qstart) > self.tend:
            # 'A' overlaps 'B'
            # a - b +
            return 4, 'L', 'L'   # should this be RL?
        else:
            # B overlaps A
            # b - a +
            return 5, 'L', 'L'



    def _self_aligned(self) -> bool:
        """
        Check if query is also target

        :return: Indicator if alignment to itself
        """
        if self.qname == self.tname:
            return True
        else:
            return False



    # def _is_merged(self) -> bool:
    #     """
    #     Check for internal match turned OVL restrictions
    #     This might be deprecated
    #
    #     :return: Indicator for already merged seqs
    #     """
    #     if '*' in self.qname and '*' in self.tname:
    #         return True
    #     else:
    #         return False



    @staticmethod
    def _is_prox(start: int, end: int, length: int, lim=1000.0) -> bool:
        """
        check if a record has a mapped region close to one of its ends
        masm definition would be lim = 1000
        a fixed basepair limit does not work for us, we need to trim off more
        i.e. if limit is given as percentage, calc a limit

        :param start: start position on seq
        :param end: end position on seq
        :param length: sequence length
        :param lim: limit of allowed overhang
        :return: indicator whether mapping in proximity to sequence end
        """
        if lim < 1:
            limit = lim * length
        else:
            limit = lim
        overhang = min(start, length - end)
        ovl = True if overhang < limit else False
        return ovl



    def _im_ovl_restrictions(self) -> bool:
        """
        check additional restrictions to count internal match as OVL

        :return: indicator for internal match to be overlap
        """
        maplen = self.map_length()
        if self.qlen > 15000:
            if self.tlen > 15000:
                if maplen > 5000:
                    return True
        return False



    def internal_match_is_overlap(self) -> bool:
        """
        due to overlapping untrimmed reads, we reconsider internal matches that might be overlaps
        if one record has a true dovetail, check if the other has a relaxed dovetail

        :return: indicator for internal match to be overlap instead
        """
        lim = 0.15
        if self._is_prox(start=self.qstart, end=self.qend, length=self.qlen):
            self.qprox = True  # mark which side the true prox is for trimming
            if self._is_prox(start=self.tstart, end=self.tend, length=self.tlen, lim=lim):
                if self._im_ovl_restrictions():
                    # check a few more restrictions
                    # relaxed dovetail
                    return True
        elif self._is_prox(start=self.tstart, end=self.tend, length=self.tlen):
            self.tprox = True  # marker for trimming
            if self._is_prox(start=self.qstart, end=self.qend, length=self.qlen, lim=lim):
                if self._im_ovl_restrictions():
                    # relaxed dovetail
                    return True
        else:
            # neither sequence has a mapping close to the end
            # i.e. true internal match
            return False
        return False



    @staticmethod
    def _find_coords(start: int, end: int, length: int) -> tuple[int, int]:
        """
        find coordinates to trim off of reads

        :param start: start of mapping on sequence to be trimmed
        :param end: end of mapping on sequence to be trimmed
        :param length: length of sequence to be trimmed
        :return: start and end coords to trim
        """
        # which side is closer to the end?
        min_overhang_idx = np.argmin([start, length - end])
        if min_overhang_idx == 0:
            # trim_start, trim_stop = start, None
            trim_start, trim_stop = 0, start
        else:
            # trim_start, trim_stop = 0, end
            trim_start, trim_stop = end, None
        return trim_start, trim_stop



    def find_trim_coords(self) -> tuple[str, int, int, str]:
        """
        if this alignment has been identified to be useful when trimmed,
        find which of the sequences we want to trim and which coordinates

        :return: Tuple of sequence name, start, end, and partner sequence
        """
        if self.qprox:
            # if q is a real prox, trim t
            sid = self.tname
            trim_start, trim_stop = self._find_coords(start=self.tstart, end=self.tend, length=self.tlen)
            other = self.qname
            other_len = self.qlen
            orig_len = self.tlen
        else:
            sid = self.qname
            trim_start, trim_stop = self._find_coords(start=self.qstart, end=self.qend, length=self.qlen)
            other = self.tname
            other_len = self.tlen
            orig_len = self.qlen

        # check that a potential trim would actually be longer than the original
        if trim_stop is None:
            stop = orig_len
        else:
            stop = trim_stop
        trimmed_bit = stop - trim_start
        # potential length of a merged sequence: original length minus trimmed bits
        # plus the length of the newly added sequence, minus the overlap length
        new_len = orig_len - trimmed_bit + other_len - self.alignment_block_length
        if new_len < orig_len:
            sid = '0'

        return sid, trim_start, trim_stop, other



    def grab_increment_coords(self) -> tuple[int, int, int, int, int, int]:
        """
        get the coordinates of a containment to increment the coverage counts

        :return: Tuple of coordinates to increment
        """
        if self.c == 2:
            ostart = self.tstart
            oend = self.tend
            cstart = self.qstart
            cend = self.qend
        elif self.c == 3:
            ostart = self.qstart
            oend = self.qend
            cstart = self.tstart
            cend = self.tend
        else:
            raise ValueError("This method should only be called on contained reads")

        olen = oend - ostart
        clen = cend - cstart
        return ostart, oend, olen, cstart, cend, clen



    def keygen(self) -> str:
        """
        New alphabetically sorted header for multiline containments

        :return: New header
        """
        if self.qname < self.tname:
            return f'{self.qname}-{self.tname}'
        else:
            return f'{self.tname}-{self.qname}'



    # for debugging
    # def plot(self, save=None):
    #     # visualize alignments as blocks
    #     import plotnine as p9
    #     import pandas as pd
    #     # this puts together the coordinates for the polygon
    #     # 5 points, starting at the lower left end
    #     if self.strand == '+':
    #         cols = ['qstart', 'qend', 'tend', 'tstart', 'qstart']
    #     else:
    #         cols = ["qend", "qstart", "tend", "tstart", "qend"]
    #
    #     pos = [getattr(self, c) for c in cols]
    #     # the query is always on top
    #     seqn = [2, 2, 1, 1, 2]
    #     cdat = pd.DataFrame({'pos': pos, 'seqn': seqn})
    #     # coordinates of the read rectangles
    #     seqlens = pd.DataFrame({'seq': [self.qname[:10], self.tname[:10]],
    #                             'start': [0, 0],
    #                             'end': [self.qlen, self.tlen],
    #                             'bottoms': [2.05, 0.8],
    #                             'tops': [2.2, 0.95]})
    #
    #     xpos = 100
    #     qpos = 1.9
    #     tpos = 1.1
    #     p = (p9.ggplot() +
    #          p9.geom_polygon(data=cdat,
    #                          mapping=p9.aes(x="pos", y="seqn"),
    #                          fill="grey", colour="black") +
    #          p9.geom_rect(data=seqlens,
    #                       mapping=p9.aes(xmin="start", xmax="end",
    #                                      ymin="bottoms", ymax="tops"),
    #                       colour="black", fill=None) +
    #          p9.annotate(x=xpos, label=self.qname[:10], y=2.12, geom='text', ha="left", va="center") +
    #          p9.annotate(x=xpos, label=self.tname[:10], y=0.87, geom='text', ha="left", va="center") +
    #          p9.annotate(x=self.qstart, label=self.qstart, y=qpos, color="darkred",
    #                      geom='text', ha="left", va="center") +
    #          p9.annotate(x=self.qend, label=self.qend, y=qpos, color="darkred",
    #                      geom='text', ha="left", va="center") +
    #          p9.annotate(x=self.tstart, label=self.tstart, y=tpos, color="darkblue",
    #                      geom='text', ha="left", va="center") +
    #          p9.annotate(x=self.tend, label=self.tend, y=tpos, color="darkblue",
    #                      geom='text', ha="left", va="center") +
    #          p9.annotate(x=xpos, label=self.strand, y=1.4, color="darkgoldenrod",
    #                      geom='text', ha="left", va="center", size=30) +
    #          p9.ylab("") +
    #          p9.xlab("") +
    #          p9.theme_minimal() +
    #          p9.theme(axis_text_y=p9.element_blank(),
    #                   axis_ticks_major_y=p9.element_blank(),
    #                   plot_background=p9.element_rect(fill="white", color="white")))
    #     if save:
    #         p.save(save)
    #     else:
    #         return p




class Paf:

    def __init__(self):
        pass


    @staticmethod
    def parse_PAF(paf_file: str | StringIO, min_len: int = 1) -> dict:
        """
        Parse the contents of a PAF file into a dictionary of records

        :param paf_file: Can be either a string or a StringIO object
        :param min_len: minimum alignment length to consider an entry
        :return: Dict of parsed PAF file
        """
        if isinstance(paf_file, str) and Path(paf_file).is_file():
            with open(paf_file, 'r') as paff:
                paf_dict = Paf._parse_content(fh=paff, min_len=min_len)
        elif isinstance(paf_file, StringIO):
            paf_dict = Paf._parse_content(fh=paf_file, min_len=min_len)
        else:
            paf_dict = dict()
            print("need file path or StringIO")
        return paf_dict



    @staticmethod
    def _parse_content(fh, min_len: int) -> dict:
        """
        Parser for PAF files into defaultdicts

        :param fh: Filehandle of PAF
        :param min_len: minimum alignment block length
        :return: parsed dict with PAF entries
        """
        paf_dict = defaultdict(list)

        for record in fh:
            paf = PafLine(record)
            # FILTERING of PAF ENTRIES
            if paf.alignment_block_length < min_len:
                continue
            if not paf.primary:
                continue

            paf_dict[str(paf.qname)].append(paf)
        return paf_dict



    @staticmethod
    def parse_filter_classify_records(paf: str, filters: SimpleNamespace) -> tuple[list, list]:
        """
        Parse a paf file by converting each line to Paf object,
        while also filtering and classifying mappings

        :param paf: String of PAF file name
        :param filters: Filters stored as Namespace
        :return: Lists of classified and skipped PAF records
        """
        records = []
        skip = []
        with open(paf, 'r') as fh:
            for record in fh:
                rec = PafLine(record)
                # check if this mapping passes the filters
                is_filtered = rec.filter(filters=filters)
                if is_filtered:
                    continue

                # classify the alignment
                rec.c = rec.classify()

                if rec.c == 1:
                    # internal match
                    skip.append(rec)
                    continue
                # if not filtered or skipped
                records.append(rec)
        return records, skip



    @staticmethod
    def choose_best_mapper(records: list) -> list:
        """
        Structured array to decide between ties, by using the score of the DP algorithm

        :param records: List of multiple mappers to decide from
        :return: Best mapper according to attributes
        """
        mapq = [(record.mapq, record.align_score) for record in records]
        custom_dtypes = [('q', int), ('dp', int)]
        mapping_qualities = np.array(mapq, dtype=custom_dtypes)
        sorted_qual = np.argsort(mapping_qualities, order=["q", "dp"])
        record = [records[sorted_qual[-1]]]
        return record



# shorthand typehint used in many places
paf_dict_type = dict[str, list[PafLine]]






import logging

import numpy as np
from numpy.typing import NDArray


class ReadlengthDist:

    def __init__(self, mu: int = 400, sd: int = 4000, lam: int = 6000, eta: int = 11):
        """
        Initializes a truncated normal distribution of read lengths.
        Used as prior for read length distribution, gets updated throughout the experiment

        :param mu: Length of anchor bases
        :param lam: Lambda parameter of the distribution.
        :param sd: Standard deviation of read lengths.
        :param eta: Number of partitions in the piece-wise approximation.
        """
        self.sd = sd
        self.lam = lam
        self.eta = eta
        self.mu = mu
        self.read_lengths = np.zeros(shape=int(1e6), dtype='uint16')
        # get the maximum read length
        longest_read = int(lam + 10 * sd)
        # prob density of normal distribution
        x = np.arange(longest_read, dtype='int')
        L = np.exp(-((x - lam + 1) ** 2) / (2 * (sd ** 2))) / (sd * np.sqrt(2 * np.pi))
        L /= sum(L)           # normalise
        self.L = L
        # stepwise approx
        self.approx_ccl = self.ccl_approx_constant()



    def update(self, read_lengths: dict) -> None:
        """
        Updates the distribution of read lengths.

        :param read_lengths: Dictionary containing read IDs and their lengths.
        """
        # loop through reads and record their length in array
        for rid, length in read_lengths.items():
            # ignore rejected reads for the read length dist
            # might overestimate the length slightly
            if length > self.mu * 2:
                # reads longer than 1M are counted as 1M
                if length >= 1e6:
                    length = int(1e6) - 1
                self.read_lengths[length] += 1
            else:
                continue

        # calc current stats of read lengths
        observed_read_lengths = np.nonzero(self.read_lengths)
        # if this function was called before any reads are observed, we just return
        if len(observed_read_lengths[0]) == 0:
            logging.info('Attempted update of read lengths before observing any reads')
            return
        length_sum = np.sum(observed_read_lengths * self.read_lengths[observed_read_lengths])
        self.lam = length_sum / np.sum(self.read_lengths[observed_read_lengths])
        self.longest_read = np.max(np.where(self.read_lengths))
        self.L = np.copy(self.read_lengths[:self.longest_read + 1]).astype('float64')
        self.L /= sum(self.L)
        self.approx_ccl = self.ccl_approx_constant()
        logging.info(f'rld: {self.approx_ccl}')
        # lambda - mu - rho
        self.time_cost = self.lam - 400 - 300



    def ccl_approx_constant(self) -> NDArray:
        """
        CCL is 1-CL (cumulative distribution) of read lengths (L).
        Starts at 1 and decreases with increasing length.
        CCL[i] is the probability that a read is at least i+1 long.
        Plus approximated by piece-wise constant function with eta pieces

        :return: approximated pieces of complementary cumulative read length distribution
        """
        # complement of cumulative distribtuion of read lengths
        ccl = np.zeros(len(self.L) + 1)
        ccl[0] = 1
        ccl[1:] = 1 - np.concatenate((self.L[1:].cumsum(), np.ones(1)))
        # cut distribution off to reduce complexity
        ccl[ccl < 1e-6] = 0
        ccl = np.concatenate((np.trim_zeros(ccl, trim='b'), np.zeros(1)))
        self.ccl = ccl
        # approx. with piecewise constant function
        approx_ccl = np.zeros(self.eta - 1, dtype='int32')
        i = 0
        for part in range(self.eta - 1):
            prob = 1 - (part + 0.5) / (self.eta - 1)
            while (ccl[i] > prob) and (len(ccl) > i):
                i += 1
            approx_ccl[part] = i
        return approx_ccl






import logging
import time
import mmap
import gzip
import pickle
from pathlib import Path
from typing import Any
from itertools import repeat
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor as TPE

import numpy as np
from numpy.typing import NDArray


# module for random read sampling in BOSS


class Sampler:

    def __init__(self, source: str, paf_full: str = None, paf_trunc: str = None, workers: int = 4, **kwargs) -> None:
        """
        Wrapper to sample sequencing data (and their mappings) given source files.
        Paf files and offsets can be generated with a script in ../scripts

        :param source: Fastq files with sequencing data to sample from
        :param paf_full: Full-length mappings of the same reads
        :param paf_trunc: Truncated (mu-sized) mappings of the same reads
        :param workers: Max number of threads for mapping retrieval
        :param kwargs: Pass-through args to the FastqStream
        """
        self.fq_stream = FastqStream_mmap(source=source, **kwargs)
        # stream for mapping data is only initialised for BOSS-RUNS, not AEONS
        self.pafs = True if paf_full and paf_trunc else False
        if self.pafs:
            self.paf_stream = PafStream(paf_full=paf_full, paf_trunc=paf_trunc, workers=workers)



    def sample(self) -> tuple[dict[str, str], dict[str, str], str, str]:
        """
        Generate a new batch of data

        :return: Tuple of read sequences, qualities and (optional) raw string output of mappings
        """
        self.fq_stream.read_batch()
        if self.pafs:
            paf_f, paf_t = self.paf_stream.grab_mappings(read_ids=self.fq_stream.read_ids)
        else:
            paf_f = defaultdict(list)
            paf_t = defaultdict(list)
        return self.fq_stream.read_sequences, self.fq_stream.read_qualities, paf_f, paf_t



class FastqStream_mmap:

    def __init__(self, source: str, seed: int = 1, shuffle: bool = False, batchsize: int = 1, maxbatch: int = 1):
        """
        Stream reads from a fastq file (4 lines per read) during simulations
        This implementation won't sample the same read twice

        :param source: Path to the file-like object.
        :param seed: Seed for random number generation.
        :param shuffle: Whether to shuffle the offsets.
        :param batchsize: Batch size.
        :param maxbatch: Maximum number of batches.
        """
        self.source = source
        if not self.source:
            raise AssertionError("No source file provided")
        # check if file is gzipped. Not very good
        suffix = source.split('.')[-1]
        if suffix == 'gz':
            self.gzipped = True
        else:
            self.gzipped = False

        self.log_each = int(int(1e5))  # log frequency of scanning for offsets
        self.filesize = int(Path(source).stat().st_size)
        logging.info(f"Representing {self.filesize / 1e6} Mbytes of data from source: {self.source}")
        # scan the offsets if the file does not exist
        if not Path(f'{source}.offsets.npy').exists():
            logging.info("scanning offsets")
            self._scan_offsets()
        logging.info("loading offsets")
        self._load_offsets(seed=seed, shuffle=shuffle, batchsize=batchsize, maxbatch=maxbatch)
        self.batch = 0



    def _scan_offsets(self, k: int = 4, limit: float = 1e9) -> None:
        """
        Scan the file to find byte offsets of each sequencing read.

        :param k: Number of lines per chunk, default chink size is 4 for fastq
        :param limit: Maximum number of reads to scan.
        :return: None.
        """
        tic = time.time()
        tmp_offsets = []
        read_num = 0

        with open(self.source, 'rb') as f:
            k_tmp = 1
            # memory-map the file; lazy eval-on-demand via POSIX filesystem
            mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            if self.gzipped:
                mm = gzip.GzipFile(mode="rb", fileobj=mm)

            for _ in iter(mm.readline, b''):
                if k_tmp % k == 0:
                    pos = mm.tell()
                    tmp_offsets.append(pos)
                    k_tmp = 1
                    read_num += 1
                    # status update
                    if read_num % self.log_each == 0:
                        logging.info(f"{read_num} reads scanned")
                else:
                    k_tmp += 1

                if read_num >= limit:
                    break

        toc = time.time()
        offsets = np.asarray(tmp_offsets, dtype='uint64')
        del tmp_offsets
        # write the offsets to a file
        np.save(f'{self.source}.offsets', offsets)
        logging.info(f"DONE scanning {read_num} reads")
        logging.info(f'wrote {len(offsets)} offsets to {self.source}.offsets.npy')
        logging.info(f"{round(toc - tic, 4)} seconds elapsed scanning file for offsets")



    def _load_offsets(self, seed: int = 1, shuffle: bool = False, batchsize: int = 1, maxbatch: int = 1) -> None:
        """
        Load offsets of sequencing reads in a fastq file.

        :param seed: Seed for random number generation.
        :param shuffle: Whether to shuffle the offsets.
        :param batchsize: Batch size.
        :param maxbatch: Maximum number of batches.
        :return: None.
        """
        if seed == 0:
            seed = np.random.randint(low=0, high=int(1e6))
        np.random.seed(seed)
        offsets = np.load(f'{self.source}.offsets.npy')
        maxbatch = maxbatch + 1

        if shuffle:
            np.random.shuffle(offsets)
            logging.info(f"offsets shuffled using random seed: {seed}")

        # shorten the offsets to number of reads we need
        len_offsets = len(offsets)
        logging.info(f"available batches: {len_offsets / batchsize}")
        n_reads = batchsize * maxbatch
        if n_reads < len_offsets:
            offsets = offsets[: n_reads]
        else:
            raise ValueError("Requested more reads than there are available in the fastq")

        # restructure the offsets into 2D array to represent batches (rows)
        offsets = offsets.reshape((maxbatch, batchsize))
        self.offsets = offsets



    def _parse_batch(self, batch_string: str) -> None:
        """
        Parse a batch in string format into dictionaries.

        :param batch_string: Batch in string format.
        :return: Tuple containing dictionaries of read lengths, sequences, sources, and the total number of bases.
        """
        read_lengths = {}
        read_sequences = {}
        read_qualities = {}
        read_sources = {}

        batch_lines = batch_string.split('\n')
        n_lines = len(batch_lines)

        i = 0
        # since we increment i by 4 (lines for read in fq), loop until n_lines - 4
        while i < (n_lines - 4):
            # grab the name of the read. split on space, take first element, trim the @
            desc = batch_lines[i].split(' ')
            name = str(desc[0][1:])
            source = str(desc[-1])
            seq = batch_lines[i + 1]
            qual = batch_lines[i + 3]
            read_len = len(seq)
            # update the containers
            read_lengths[name] = read_len
            read_sequences[name] = seq
            read_sources[name] = source
            read_qualities[name] = qual
            i += 4
        # get the total length of reads in this batch
        total_bases = int(np.sum(np.array(list(read_lengths.values()))))
        # assign attributes
        self.read_ids = set(read_sequences.keys())
        self.read_lengths = read_lengths
        self.read_sequences = read_sequences
        self.read_qualities = read_qualities
        self.read_sources = read_sources
        self.total_bases = total_bases



    @staticmethod
    def _get_single_read(mm: Any, offset: int) -> str:
        """
        Return 4 lines from a memory-mapped fastq file given a byte-wise position.

        :param mm: The memory-mapped fastq file.
        :param offset: The byte-wise position of the read.
        :return: The read chunk.
        """
        mm.seek(offset)
        chunk_size = 4
        chunk = b''
        # read 4 lines of the fastq entry
        for _ in range(chunk_size):
            chunk += mm.readline()
        chunk = chunk.decode("utf-8")
        return chunk



    def read_batch(self, delete: bool = True) -> None:
        """
        Return a batch of reads from the fastq file.

        :param delete: Whether to delete the batch offsets after retrieval.
        :return:
        """
        # check if offsets are empty
        if self.offsets.shape[0] == 0:
            raise ValueError("No more reads left to sample. Exiting")

        with open(self.source, 'rb') as f:
            mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            if self.gzipped:
                mm = gzip.GzipFile(mode="rb", fileobj=mm)
            # the first row of the offsets are the next batch
            batch_offsets = self.offsets[0, :]
            # initialise list instead of string concat
            batch = [''] * len(batch_offsets)
            # possibly LINUX specific and not POSIX
            # here we tell the kernel to preload pages of the mapped memory
            # magically makes it even faster to access
            # pagesize is a LINUX (system)-specific constant of 4096 bytes per "page"
            pagesize = 4096
            # the start of WILLNEED needs to be a multiple of pagesize
            # we take the modulo and move the start of the offset a bit earlier if needed
            new_offsets = batch_offsets - (batch_offsets % pagesize)

            if not self.gzipped:
                for new_offset in new_offsets:
                    # we preload 20 pages of data following each read start
                    # 20 pages = 80 kbytes (read of up to ~40 kbases, I think)
                    mm.madvise(mmap.MADV_RANDOM)
                    mm.madvise(mmap.MADV_WILLNEED, int(new_offset), 20)

            batch_offsets = np.sort(batch_offsets)
            for i in range(len(batch_offsets)):
                try:
                    # jump to position in file and return the next 4 lines
                    chunk = self._get_single_read(mm=mm, offset=int(batch_offsets[i]))
                    batch[i] = chunk
                except:
                    logging.info(f"Error at location: {batch_offsets[i]}")
                    continue
                if len(chunk) == 0:
                    continue

            # add call to close memory map, only file itself is under with()
            mm.close()

        if not batch[0].startswith('@') and not batch[0].startswith('>'):
            raise ValueError("The batch of reads is broken")

        if delete:
            # remove the row from the offsets, so it does not get sampled again
            new_offsets = np.delete(self.offsets, 0, 0)
            self.offsets = new_offsets
        # parse the batch (long string) into dicts
        batch_string = ''.join(batch)
        self._parse_batch(batch_string=batch_string)
        self.batch += 1
        logging.info(f'got new batch of {len(self.read_sequences)} reads')





class PafStream:


    def __init__(self, paf_full: str, paf_trunc: str, workers: int = 4):
        """
        Initialise a PafStream object to quickly grab paf entries given sampled read IDs

        :param paf_full: Path to paf file with mappings of full-length reads
        :param paf_trunc: Path to paf file with mappings of truncated reads
        :param workers: Max threads to spawn
        """

        self.paf_full = paf_full
        self.paf_trunc = paf_trunc
        self.workers = workers

        # scan the offsets if the file does not exist
        if not Path(f'{paf_full}.offsets').exists():
            logging.info("scanning full-length PAF offsets")
            self._scan_offsets_paf(path=self.paf_full)
        self.offsets_full = self._load_offsets_paf(path=paf_full)

        # same for truncated mappings
        if not Path(f'{paf_trunc}.offsets').exists():
            logging.info("scanning truncated PAF offsets")
            self._scan_offsets_paf(path=self.paf_trunc)
        self.offsets_trunc = self._load_offsets_paf(path=paf_trunc)
        self.batch = 0



    @staticmethod
    def _load_offsets_paf(path: str) -> defaultdict[list]:
        """
        Load offsets from file

        :param path: Path to offsets file
        :return: Defaultdict with list of mappings of reads
        """
        # check if offsets present
        if not Path(f'{path}.offsets').is_file():
            print("no paf offsets found. exiting")
            exit(1)

        with open(f'{path}.offsets', "rb") as p:
            offsets = pickle.load(p)
        logging.info(f"loaded offsets for {path.split('/')[-1]}")
        return offsets



    @staticmethod
    def _scan_offsets_paf(path: str, limit: int = 1e9) -> None:
        """
        Scan file to find byte offsets of mappings

        :param path: Path to paf mapping file
        :param limit: maximum lines to scan (debugging)
        :return:
        """
        mapping_num = 0
        offsets = defaultdict(list)
        pos = 0

        with open(path, 'rb') as paf:
            for line in paf:
                ll = line.split(b"\t")
                read_id = ll[0].decode()
                offsets[read_id].append(pos)
                pos = paf.tell()
                mapping_num += 1

                if mapping_num % 1e5 == 0:
                    print(f"{mapping_num} mappings scanned")

                if mapping_num >= limit:
                    break

        # write the offsets to a file
        with open(f'{path}.offsets', 'wb') as p:
            pickle.dump(offsets, p)
        print(f"DONE scanning {mapping_num} mapping records")



    @staticmethod
    def _grab_paf_lines(args: tuple[str, NDArray]) -> str:
        """
        Helper func for multithreaded grabbing of paf lines

        :param args: tuple of file path and array of byte offsets
        :return:
        """
        path, pos = args
        paf = b''
        with open(path, 'rb') as f:
            for p in pos:
                f.seek(p)  # seek to position
                paf += f.readline()
        return paf.decode()



    @staticmethod
    def _get_offset_array(offsets: defaultdict[list], read_ids: set) -> NDArray:
        """
        Given a set of read IDs, get the array of offsets of the mappings of those reads

        :param offsets: Defaultdict of offsets of mappings
        :param read_ids: Set of reads from which to retrieve offsets
        :return: Array of offsets of mappings of requested reads
        """
        rid_offsets = [offsets[rid] for rid in read_ids]
        unmapped = np.sum([1 for i in rid_offsets if len(i) == 0])
        mapped = np.sum([1 for i in rid_offsets if len(i) > 0])
        logging.info(f"mapped {mapped}, unmapped {unmapped}")
        # transform offsets into array
        offsets_arr = np.concatenate([np.array(i, dtype=int) for i in rid_offsets])
        return offsets_arr


    def _grab_pafs(self, offsets: NDArray, path: str) -> str:
        """
        Multithreaded retrieval of paf mappings given an array of offsets

        :param offsets: Array of offsets of a specific set of reads
        :param path: Path to the original file to grab from
        :return: Concatenated mappings of interest
        """
        worker_offsets = np.array_split(offsets, self.workers)
        with TPE(max_workers=self.workers) as executor:
            pafs = executor.map(self._grab_paf_lines,
                                zip(repeat(path),
                                    worker_offsets))
        paf_out = ''.join(pafs)
        return paf_out


    def grab_mappings(self, read_ids: set) -> tuple[str, str]:
        """
        Given a set of read IDs, get their offsets and grab the mappings data from these reads

        :param read_ids: Set of read IDs
        :return: Tuple of mappings for full and truncated mappings of the reads
        """
        # given some sampled read ids, grab their full and trunc mappings
        offsets_arr_full = self._get_offset_array(
            offsets=self.offsets_full, read_ids=read_ids)
        offsets_arr_trunc = self._get_offset_array(
            offsets=self.offsets_trunc, read_ids=read_ids)
        # grab the actual data
        paf_f = self._grab_pafs(
            offsets=offsets_arr_full, path=self.paf_full)
        paf_t = self._grab_pafs(
            offsets=offsets_arr_trunc, path=self.paf_trunc)
        return paf_f, paf_t




import sys
import os
from pathlib import Path
import time
import argparse

import numpy as np
import mappy

from boss._cli_base import main as main_args

from readfish._config import Conf  # noqa
from readfish.plugins.abc import AlignerABC
from readfish.plugins.utils import Decision, Result



class BossBits:

    def __init__(self, conf, logger, mapper):
        # initialise the bits that are necessary for dynamic updates
        self.scale_factor = 100  # make param in future maybe
        self.masks = {}
        self.logger = logger
        self.mapper = mapper
        self.last_mask_mtime = 0
        self.last_contig_mtime = 0

        # Overwrite the default strand_converter
        # used to index into the strategies,
        # which are shaped 2xN for forward & reverse in 1st dim
        # readfish uses 1 for forward, -1 for reverse
        # For BOSS we use 0/False for forward, 1/True for reverse
        self.strand_converter = {1: False, -1: True}

        # grab path to masks and contigs, requires presence of 1 BOSS region in toml
        for region in conf.regions:
            # get the regions that is not the control
            # this currently limits using BOSS in a single region
            if region.control is False:
                self.mask_path = Path(f"out_{region.name}/masks")
                self.cont_path = Path(f"out_{region.name}/contigs")
                if not self.mask_path.is_dir():
                    os.makedirs(self.mask_path)
                    os.makedirs(self.cont_path)

        self.logger.info(f"Looking for BOSS related files at: \n"
                         f" strategies: {self.mask_path} \n"
                         f" contigs: {self.cont_path}")
        # if there is no strategy, generate dummy to start with
        if not Path(self.mask_path / 'boss.npz').is_file():
            self.logger.info("Creating dummy strategy")
            contig_strats = {'init': np.ones(1)}
            np.savez(self.mask_path / "boss", **contig_strats)
        else:
            self.logger.info("Loading starting strategy")


    @staticmethod
    def _reload_npz(mask_files):
        mask_container = np.load(mask_files[0])
        return {name: mask_container[name] for name in mask_container}


    @staticmethod
    def _idx(fa):
        idx_name = fa.stem + ".mmi"
        mappy.Aligner(fn_idx_in=str(fa), fn_idx_out=str(idx_name), preset="map-ont")
        return idx_name


    def _reload_masks(self):
        """
        Reload updated decision masks.
        """
        # BOSS* uses a single .npz in all modes now
        new_masks = list(self.mask_path.glob("boss.npz"))
        reload_func = self._reload_npz
        if not new_masks:
            raise FileNotFoundError("No mask files present")

        # Do we actually update this time?
        if not new_masks[0].stat().st_mtime > self.last_mask_mtime:
            return 0

        try:
            mask_dict = reload_func(mask_files=new_masks)
            self.masks = mask_dict
            self.logger.info(f"Reloaded strategies for {len(set(self.masks.keys()))} sequences")
        except Exception as e:
            self.logger.error(f"Error reading strategy array ->>> {repr(e)}")
            self.masks = {"exception": True}
        # update last mtime
        self.last_mask_mtime = new_masks[0].stat().st_mtime
        return 1


    def _reload_mapper(self, conf):
        """
        Reload mapping index for decision-making. Only load if marker file exists.
        """
        # Do we actually update this time?
        # check if contigs are newer than before
        cs = self.cont_path / "aeons.fa"

        if not cs.stat().st_mtime > self.last_contig_mtime:
            return

        try:
            self.logger.info(f"Regenerating mapping index")
            idx_name = self._idx(fa=cs)
            self.logger.info(f"Reloading mapping index")
            conf.mapper_settings.parameters['fn_idx_in'] = idx_name
            mapper: AlignerABC = conf.mapper_settings.load_object("Aligner")
            self.mapper = mapper
            self.logger.info("Aligner initialised")
            while not self.mapper.initialised:
                time.sleep(1)
            # new last mtime
            self.last_contig_mtime = cs.stat().st_mtime

        except Exception as e:
            self.logger.error(f"Error loading mapping index ->>> {repr(e)}")


    def _check_names(self):
        """
        Check that the names of masks and sequences in the mapper are the same.
        """
        if not self.mapper.initialised:
            return

        mask_names = sorted(list(self.masks.keys()))
        contig_names = sorted(list(self.mapper.aligner.seq_names))
        same_names = mask_names == contig_names
        if not same_names:
            self.logger.error(f"Error loading masks and contigs: discrepancy in names {len(mask_names)} {len(contig_names)} \n\n\n\n\n")


    def reload(self, conf):
        # masks are reloaded from mask files
        reloaded_masks = self._reload_masks()
        if not reloaded_masks:
            return 0, self.mapper
        # return if running BOSS-RUNS
        if not (self.cont_path / "aeons.fa").exists():
            return 0, self.mapper
        self._reload_mapper(conf)
        # check that the names of the masks and contigs are the same
        self._check_names()
        return 1, self.mapper


    def _check_coord(self, contig, start_pos, reverse):
        """
        Query numpy mask array and return decision to keep sequencing
        Parameters
        ----------
        start_pos: int
            Mapping start coordinate
        reverse: bool
            Whether the coordinate is on the forward or reverse strand
        contig: str
            The name of the contig the mask is for, used to lookup correct mask in the directory
        Returns
        -------
        bool
            Decision to keep sequencing
        """
        # this happens when there was an error reading the mask file, then accept everything
        if self.masks.get("exception", False):
            return 1
        if contig not in self.masks:
            self.logger.warning(f"{contig} is not in mask dict")
            return 1
        arr = self.masks.get(contig)
        # reject refs have array size 1
        if arr.shape[0] == 1:
            return 0
        # otherwise query the strategy array
        try:
            d = arr[:, int(reverse)][start_pos // self.scale_factor]
            return d
        except Exception as e:  # noqa
            return 1


    def make_decision_boss(self, conf: Conf, result: Result) -> Decision:  # noqa
        """
        The main decision-making function for readfish.
        Modified from _config.py in readfish
        Chooses the decision that is looked up in the TOML based
        on the mapped coordinates of the read, checked against the targets.
        Decision is one of single_on, multi_on, single_off, multi_off, no_map, no_seq.

        :param conf: The Conf object for the experiment
        :param result: The result of the alignment and base calling for the read
        :raises ValueError: If readfish fails to make a decision based on
            the passed Result object length
        :return: The decision that readfish has made for this read
        """
        if result.alignment_data is None:
            result.alignment_data = []
        # targets = conf.get_targets(result.channel, result.barcode)
        results = result.alignment_data
        matches = []
        for al in results:
            contig = al.ctg
            strand = al.strand  # noqa
            coord = al.r_st if al.strand == 1 else al.r_en
            # matches.append(targets.check_coord(contig, strand, coord))
            strand_conv = self.strand_converter[al.strand]
            matches.append(self._check_coord(contig=contig, start_pos=coord, reverse=strand_conv))
        coord_match = any(matches)

        if not results:
            if len(result.seq) > 0:
                return Decision.no_map
            else:
                return Decision.no_seq
        elif len(results) == 1:
            return Decision.single_on if coord_match else Decision.single_off
        elif len(results) > 1:
            return Decision.multi_on if coord_match else Decision.multi_off
        raise ValueError()


    @staticmethod
    def gen_dummy_idx() -> None:
        """
        Generate a dummy index for readfish to start out with

        :return:
        """
        # generate dummy index
        with open("dummy.fa", "w") as dfa:
            dfa.write(">init\nAAAAAAAAAAAAAAAAAAAAAAAAA")
        mappy.Aligner(fn_idx_in="dummy.fa", fn_idx_out="readfish.mmi", preset="map-ont")
        Path("dummy.fa").unlink()




def get_args(arg_list: list = None) -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    """
    Get the default arguments and extend with toml file using _cli_base.main()
    :param arg_list: List to pass in arguments for debugging
    :return: parser and arguments to pass to the entry-point
    """
    if not arg_list:
        toml_readfish = sys.argv[1]
        device = sys.argv[2]
        name = sys.argv[3]
        # host = sys.argv[3]
        # port = sys.argv[4]
    else:
        toml_readfish = arg_list[0]
        device = arg_list[1]
        name = arg_list[2]

    # check arguments to add in future with: readfish targets -h
    argv = [
        'targets',
        '--toml', toml_readfish,
        '--device', device,
        '--experiment-name', name,
        # '--log-file', 'readfish.log'
        # '--host', host,
        # '--port', port,
    ]
    if arg_list:
        argv.append("--debug-log")
        argv.append("readfish_chunks.tsv")
    parser, args = main_args(argv=argv)
    return parser, args





from pathlib import Path
import numpy as np
import sys


DIR = Path(sys.argv[1])
DIR_C = DIR / "contigs"
DIR_M = DIR / "masks"



def random_id(k: int = 20) -> str:
    """
    Generate a random alphanumeric ID of length k.

    :param k: The length of the ID.
    :type k: int
    :return: The generated random ID.
    :rtype: str
    """
    import random
    import string
    x = ''.join(random.choices(string.ascii_letters + string.digits, k=k))
    return x


def gen_dummy_seqs(nseqs):
    # fill a dictionary with n random contigs
    seqs = {}
    for i in range(nseqs):
        i_id = random_id()
        i_len = np.random.randint(1000, 10000, size=1)
        i_seqlist = np.random.choice(['A', 'C', 'G', 'T'], size=i_len, replace=True)
        i_seq = "".join(i_seqlist)
        seqs[i_id] = i_seq
    return seqs



def seqs_to_file(seqs):
    with open(f"{DIR_C}/aeons.fa", "w") as dfa:
        for header, seq in seqs.items():
            dfa.write(f'>{header}\n{seq}\n')



def gen_dummy_strat(seqs):
    contig_strats = {}
    for header, seq in seqs.items():
        random_strat = np.random.choice([0, 1], size=(len(seq), 2), replace=True)
        contig_strats[header] = random_strat

    np.savez(DIR_M / "boss.npz", **contig_strats)



sequences = gen_dummy_seqs(5)
seqs_to_file(sequences)
gen_dummy_strat(sequences)





import mappy
import sys

# fasta file is read as positional argument
ref = sys.argv[1]
# name for index file
idx = f"{ref}.mmi"
# calling mappy with fn_idx_out writes index to file
aligner = mappy.Aligner(ref, preset="map-ont", n_threads=24, fn_idx_out=idx)



from time import sleep

import boss.config
from boss.runs import core as br_core
from boss.runs import simulation as br_sim
from boss.aeons import core as ba_core
from boss.aeons import simulation as ba_sim





def main(arg_list: list = None):
    # load arguments
    # DEBUG
    # arg_list = [
    # '--toml', "./config/BOSS_AEONS_SIM.toml"
    # ]
    arg_list = None if not arg_list else arg_list
    conf = boss.config.Config(parse=True, arg_list=arg_list)
    args = conf.args

    # launch entry-point
    # Real experiment mode
    if args.live_run:
        # initialise mode
        if args.ref:
            # runs initialises then launches readfish
            exp = br_core.BossRuns(args=args)
            exp.init()
            exp.launch_live_components()
            func = exp.process_batch_runs
        else:
            # aoens needs readfish running to have channel numbers available
            exp = ba_core.BossAeons(args=args)
            exp.launch_live_components()
            exp.init()
            func = exp.process_batch_aeons
        # launch main loop
        try:
            while True:
                next_update = exp.process_batch(func)
                # wait until next update
                if next_update > 0:
                    sleep(next_update)

        except KeyboardInterrupt:
            print("exiting after keyboard interrupt.. ")


    # Simulation mode
    else:
        # initialise mode
        if args.ref:
            exp = br_sim.BossRunsSim(args=args)
            func = exp.process_batch_runs_sim
        else:
            exp = ba_sim.BossAeonsSim(args=args)
            func = exp.process_batch_aeons_sim
        # launch main loop
        exp.init_sim()
        while exp.batch <= args.maxb:
            exp.process_batch_sim(func)
        exp.cleanup()



if __name__ == "__main__":
    main()





import logging
import re
from collections import defaultdict
from itertools import permutations, product
from concurrent.futures import ThreadPoolExecutor as TPexe
from typing import Any

import numpy as np
from numpy.typing import NDArray
from numba import njit

from boss.utils import reverse_complement, binc, adjust_length
from boss.paf import Paf, paf_dict_type



class Priors:

    def __init__(self, ploidy: int = 1):
        """
        Initialise a prior container

        :param ploidy: Only haploid and diploid implemented at the moment
        """
        # set ploidy to either haploid or diploid
        if int(ploidy) == 1:
            self.diploid = False
        elif int(ploidy) == 2:
            self.diploid = True
        else:
            raise ValueError("Given ploidy is not defined")
        # initialise phi and priors
        self.len_b, self.len_g, self.phi = self._generate_phi(diploid=self.diploid)
        self._init_phi_stored()
        self._init_prior()




    @staticmethod
    def _generate_phi(
        diploid: bool = False,
        deletion_error: float = 0.03,
        err_missed_deletion: float = 0.1,
        substitution_error: float = 0.04,
    ) -> tuple[int, int, NDArray]:
        """
        Generate phi. For haploid & diploid. Either with deletions as genotype or ignoring them

        :param diploid: diploid or haploid
        :param deletion_error: probability that a base in the genome is skipped in a read (observed - while in truth it's not)
        :param err_missed_deletion: Probability that an actual deletion is missed in a read. e+
        :param substitution_error: Probability that wrong base is read.
        :return: Tuple of number of bases, genotypes, and phi array
        """
        # two distinctions: haploid/diploid & deletions
        if not diploid:
            if not deletion_error:
                # simplest case: haploid, no deletion genotype
                bases = ['A', 'C', 'G', 'T']
                genotypes = bases
                lenB = len(bases)
                lenG = len(genotypes)
                phi = np.zeros((lenB, lenG))
                for b in range(lenB):
                    for g in range(lenG):
                        if b == g:
                            phi[b][g] = 1 - substitution_error
                        else:
                            phi[b][g] = substitution_error / (lenB - 1)

            else:
                # haploid but with deletions
                bases = ["A", "C", "G", "T", "-"]
                genotypes = bases
                lenB = len(bases)
                lenG = len(genotypes)
                phi = np.zeros((lenB, lenG))
                for b in range(lenB - 1):
                    for g in range(lenG - 1):
                        if b == g:
                            phi[b][g] = 1.0 - (substitution_error + deletion_error)
                        else:
                            phi[b][g] = substitution_error / (lenB - 2)

                        # prob that a base is true when deletion is detected
                        phi[lenB - 1][g] = deletion_error

                    # prob that a deletion is true when base is detected
                    phi[b][lenG - 1] = err_missed_deletion / (lenB - 1)

                # prob that deletion is true when deletion is detected
                phi[lenB - 1][lenG - 1] = 1.0 - err_missed_deletion

        else:
            if not deletion_error:
                # diploid, no deletions
                bases = 'ACGT'
                genotypes = ['AA', 'AC', 'AG', 'AT', 'CC', 'CG', 'CT', 'GG', 'GT', 'TT']
                lenB = len(bases)
                lenG = len(genotypes)
                phi = np.zeros((lenB, lenG))
                for b in range(lenB):
                    for g in range(lenG):
                        # countBase indicates homozygous, heterozygous or diff to reference
                        countBase = genotypes[g].count(bases[b])
                        if countBase == 2:
                            phi[b][g] = 1 - substitution_error
                        elif countBase == 1:
                            phi[b][g] = ((1 - substitution_error) / 2) + (substitution_error / (2 * (lenB - 1)))
                        elif countBase == 0:
                            phi[b][g] = substitution_error / (lenB - 1)

            else:
                # most complex phi: diploid, with deletions
                bases = 'ACGT-'
                genotypes = ['AA', 'AC', 'AG', 'AT', 'CC', 'CG', 'CT', 'GG', 'GT', 'TT',
                             'A-', 'C-', 'G-', 'T-', '--']
                lenB = len(bases)
                lenG = len(genotypes)
                phi = np.zeros((lenB, lenG))
                prob_no_error = 1.0 - (substitution_error + deletion_error)
                for b in range(lenB - 1):
                    for g in range(lenG - 5):
                        # the standard base genotypes
                        countBase = genotypes[g].count(bases[b])
                        if countBase == 2:
                            phi[b][g] = prob_no_error  # homozygous
                        elif countBase == 1:
                            phi[b][g] = prob_no_error / 2 + substitution_error / (2 * (lenB - 2))  # heterozygous
                        elif countBase == 0:
                            phi[b][g] = substitution_error / (lenB - 2)  # total difference

                    # treatment of deletion genotypes
                    for g in range(10, 14):
                        countBase = genotypes[g].count(bases[b])
                        if countBase == 1:
                            # heterozygous del if base true
                            phi[b][g] = prob_no_error / 2 + err_missed_deletion / (2 * (lenB - 1))
                        elif countBase == 0:
                            # heterozygous del if base false
                            phi[b][g] = substitution_error / (2 * (lenB - 2)) + err_missed_deletion / (2 * (lenB - 1))

                    # homozygous del if base actually true
                    phi[b][lenG - 1] = err_missed_deletion / (lenB - 1)

                # probs when deletion is truth
                for g in range(lenG):
                    countGap = genotypes[g].count(bases[lenB - 1])
                    if countGap == 2:
                        phi[lenB - 1][g] = 1.0 - err_missed_deletion  # homozygous del
                    elif countGap == 1:
                        phi[lenB - 1][g] = (1.0 - err_missed_deletion) / 2 + deletion_error / 2  # heterozygous del
                    elif countGap == 0:
                        phi[lenB - 1][g] = deletion_error  # no del observed

        return lenB, lenG, phi



    def _init_phi_stored(self) -> None:
        """
        initialise the stored version of phi
        :return:
        """
        phi_stored = np.full((self.len_b, self.len_g, 1000), 1.0)
        for i in range(self.len_b):
            for j in range(self.len_g):
                phi_stored[i, j, :] = self.phi[i, j] ** np.arange(1000)
        self.phi_stored = phi_stored



    def _init_prior(self) -> None:
        """
        Initialize matrix of priors for all possible genotypes
        :return:
        """
        if not self.diploid:
            self.priors = self._haploid_priors()
        else:
            self.priors = self._diploid_priors()
        self.prior_dist = np.array([self.priors[0]])



    @staticmethod
    def _haploid_priors(
        deletion_error: float = 0.03,
        theta: float = 0.01,
        del_subs_ratio: float = 0.4
    ) -> NDArray:
        """
        Initialise priors for haploid genotypes

        :param deletion_error: probability that a base in the genome is skipped in a read (observed - while in truth it's not)
        :param theta: Population diversity
        :param del_subs_ratio: Ratio of deletions and substitutions
        :return:
        """
        if not deletion_error:
            # haploid, no deletions
            bases = ['A', 'C', 'G', 'T']
            genotypes = bases
            len_b = len(bases)
            len_g = len(genotypes)
            priors = np.zeros((len_b, len_g))

            for i in range(len_b):
                for j in range(len_g):
                    if i == j:
                        # prior for observing ref
                        priors[i][j] = 1.0 - theta
                    else:
                        # prior for observing base different to ref
                        priors[i][j] = theta / 3


        else:
            # haploid with deletions
            bases = ["A", "C", "G", "T", "-"]
            genotypes = bases
            len_b = len(bases)
            len_g = len(genotypes)
            priors = np.zeros((len_b - 1, len_g))

            for i in range(len_b - 1):
                for j in range(len_g - 1):
                    if i == j:
                        # prior for observing ref
                        priors[i][j] = 1.0 - (theta * (1.0 + del_subs_ratio))
                    else:
                        # prior for observing base different to ref
                        priors[i][j] = theta / 3

            # prior for observing a deletion
            if del_subs_ratio > 0.0001:
                priors[:, -1] = theta * del_subs_ratio
        return priors



    @staticmethod
    def _diploid_priors(
        deletion_error: float = 0.03,
        theta: float = 0.01,
        del_subs_ratio: float = 0.4
    ) -> NDArray:
        """
        Initialise priors for diploid genotypes

        :param deletion_error: probability that a base in the genome is skipped in a read (observed - while in truth it's not)
        :param theta: Population diversity
        :param del_subs_ratio: Ratio of deletions and substitutions
        :return:
        """
        # The prior proportion of diploid genome sites different from the reference expected to be homozygous
        # depends only on population size
        popsize = 1000
        homo = 0.0
        hetero = 0.0
        aN = np.sum(1.0 / (np.arange(1, popsize + 1)))
        for i in range(popsize):
            homo += (1.0 / ((i + 1) * aN)) * ((i + 1) * float(i + 1) / (popsize ** 2))
            hetero += (1.0 / ((i + 1) * aN)) * 2 * ((popsize - (i + 1)) * float(i + 1) / (popsize ** 2))
        p_homo = homo / (homo + hetero)

        if not deletion_error:
            # diploid, no deletions
            bases = 'ACGT'
            genotypes = ['AA', 'AC', 'AG', 'AT', 'CC', 'CG', 'CT', 'GG', 'GT', 'TT']
            len_b = len(bases)
            len_g = len(genotypes)
            priors = np.zeros((len_b, len_g))

            for b in range(len_b):
                for g in range(len_g):
                    # homo- or heterozygous observations in g compared to ref b
                    countBase = genotypes[g].count(bases[b])
                    if countBase == 2:
                        priors[b][g] = 1 - theta
                    elif countBase == 1:
                        priors[b][g] = ((1 - p_homo) * theta) / 3
                    elif countBase == 0:
                        priors[b][g] = (p_homo * theta) / 3

        else:
            # diploid with deletions
            bases = 'ACGT-'
            genotypes = ['AA', 'AC', 'AG', 'AT', 'CC', 'CG', 'CT', 'GG', 'GT', 'TT',
                         'A-', 'C-', 'G-', 'T-', '--']
            len_b = len(bases)
            len_g = len(genotypes)
            priors = np.zeros((len_b - 1, len_g))

            for b in range(len_b - 1):
                for g in range(len_g - 5):
                    # standard base situation
                    countBase = genotypes[g].count(bases[b])
                    if countBase == 2:
                        priors[b][g] = 1 - theta * (1 + del_subs_ratio)  # homozygous
                    elif countBase == 1:
                        priors[b][g] = ((1 - p_homo) * theta) / 3  # heterozygous
                    elif countBase == 0:
                        priors[b][g] = (p_homo * theta) / 3  # total difference

                # treatment of deletion genotypes
                for g in range(10, 14):
                    # prior for heterozygous deletion
                    # should we have a distinction between g2 = b_r and g2 != b_r?
                    priors[b][g] = (1 - p_homo) * del_subs_ratio * theta

                # prior for homozygous deletion
                priors[b][len_g - 1] = p_homo * del_subs_ratio * theta
        return priors




    def uniform_priors(self) -> None:
        """
        change the priors of genotype probabilities to be uniform
        :return:
        """
        len_g = self.priors.shape[1]
        uniform_p = 1 / len_g
        self.priors.fill(uniform_p)
        self.prior_dist = np.array([self.priors[0]])






class Scoring:

    def __init__(self, ploidy: int = 1):
        """
        Initialise a scoring container
        :param ploidy: Ploidy of organism in experiment
        """
        self.priors = Priors(ploidy=ploidy)
        self.n_ref = 4  # n of distinct reference nuc
        self.score0, self.ent0 = self.calc_score(scores=np.zeros(1), pos_posterior=self.priors.prior_dist[0:1])




    def init_score_array(self, max_cov: int = 40, size: int = 4) -> None:
        """
        Function to precompute the posterior and scores
        of different coverage patterns. The pattern generating code uses
        itertools.product() to produce all possible combinations of the ranges
        (all orderings, with repeated elements), then all permutations of the
        combinations are made with itertools.permutations()
        (all possible orderings, no repeats). This makes sure the patterns
        are produced for all possible reference bases.
        Posteriors and scores are then calculated and placed in their locations of
        a multidimensional array for fast indexing.

        :param max_cov: Max size of precomputed array
        :param size: defines sparsity of precomputed array
        :return:
        """
        # create cartesian product
        # if deletions are not accounted for, array will be the same but deletion dimension will not
        # influence the score
        p = np.array(
            list(
                product(range(int(max_cov)),
                        range(int(max_cov / size)),
                        range(int(max_cov / (size + 7))),
                        range(int(max_cov / (size + 9))),
                        range(int(max_cov / (size + 11))))
                ), dtype='uint8'
        )
        # all permutations per element
        cov_patterns = set()
        for i in range(len(p)):
            q = set(list(permutations(p[i])))
            cov_patterns |= q

        # transform generated coverage patterns to array
        cov_patterns = np.array(list(cov_patterns), dtype='uint16')
        n = len(cov_patterns)
        # calculate posterior and scores for coverage patterns
        entropies, scores = self.calc_posterior_and_scores(cov_patterns=cov_patterns)
        # construct lookup table with multidimensional array
        self.score_arr = np.zeros(shape=(max_cov, max_cov, max_cov, max_cov, max_cov, self.n_ref))
        self.entropy_arr = np.zeros(shape=(max_cov, max_cov, max_cov, max_cov, max_cov, self.n_ref))
        ind = np.swapaxes(cov_patterns, 0, 1)
        # assign scores and entropies to their locations
        for i in range(n):
            self.score_arr[ind[0, i], ind[1, i], ind[2, i], ind[3, i], ind[4, i]] = scores[:, i]
            self.entropy_arr[ind[0, i], ind[1, i], ind[2, i], ind[3, i], ind[4, i]] = entropies[:, i]




    def update_scores(self, contig) -> tuple[NDArray, NDArray]:
        """
        Update the scores of a contig and return the updated scores
        This is a method of the scoring class to make the large score array persistent without reassigning

        :param contig: Contig object for which to update scores
        :return: Arrays of scores and entropy to reassign
        """
        # grab the correct arrays
        scores = contig.scores
        entropy = contig.entropy
        coverage = contig.coverage
        # this is run after coverage has been updated
        cm = contig.change_mask
        # set deletions to 0 if not counting them
        if contig.len_b == 4:
            coverage[:, 4] = 0
        # indices where array_limit is reached
        maxed_indices = np.where(coverage.sum(axis=1) >= 30)[0]
        cm[maxed_indices] = False
        # indices of change
        cc_pos = np.nonzero(cm)[0]
        # target coverage patterns
        cc = coverage[cc_pos]
        # grab base of changed positions
        ref_bases = contig.seq_int[cc_pos]
        # main operation: indexing in pre-computed values using coverage patterns
        scores[cc_pos] = self.score_arr[cc[:, 0], cc[:, 1], cc[:, 2], cc[:, 3], cc[:, 4], ref_bases]
        # prevent maxed sites from being recalculated
        scores[maxed_indices] = np.finfo(float).tiny
        # check positions where scores were not found in the scoreArray
        # i.e. if they remain 0.0
        missing = np.argwhere(scores == 0.0).flatten()
        nmiss = missing.shape[0]
        # if any positions without scores, calc and add to array
        if nmiss != 0:
            missing_patterns = coverage[missing]
            # calculate new scores and entropies
            miss_entropies, miss_scores = self.calc_posterior_and_scores(cov_patterns=missing_patterns)
            # place into pre-computed array
            bases = contig.seq_int[missing]
            scores[missing] = miss_scores[bases, np.arange(nmiss)]
            entropy[missing] = miss_entropies[bases, np.arange(nmiss)]
            ind = np.swapaxes(missing_patterns, 0, 1)
            # assign the calculated scores and entropies to the large array
            for i in range(ind.shape[1]):
                self.score_arr[ind[0, i], ind[1, i], ind[2, i], ind[3, i], ind[4, i]] = miss_scores[:, i]
                self.entropy_arr[ind[0, i], ind[1, i], ind[2, i], ind[3, i], ind[4, i]] = miss_entropies[:, i]
        # grab the entropy values of the changed sites
        entropy[cc_pos] = self.entropy_arr[cc[:, 0], cc[:, 1], cc[:, 2], cc[:, 3], cc[:, 4], ref_bases]
        return scores, entropy




    def calc_posterior_and_scores(self, cov_patterns: NDArray) -> tuple[NDArray, NDArray]:
        """
        Calculate posterior and positional benefits. Operates on any
        number of given coverage patterns and return the values for all
        possible reference bases. Used for initialising the score_arr
        and to calculate missing values.

        :param cov_patterns: Array with different patterns of coverage for which to calc the posterior and scores
        :return: tuple of entropy and scores of input patterns
        """
        n = len(cov_patterns)
        # container for scores with same length as cov_patterns
        scores = np.repeat(self.score0, repeats=n, axis=0)
        pattern_posterior = self.calc_posterior(target_cov=cov_patterns)
        # calculate scores from posteriors
        pattern_score = np.zeros((self.n_ref, n))
        pattern_entropy = np.zeros((self.n_ref, n))
        for i in range(self.n_ref):
            tmpScore, tmpEnt = self.calc_score(scores=scores, pos_posterior=pattern_posterior[i, :, :])
            pattern_score[i] = tmpScore
            pattern_entropy[i] = tmpEnt
        return pattern_entropy, pattern_score



    def calc_posterior(self, target_cov: NDArray) -> NDArray:
        """
        Calculate posterior for site patterns.

        :param target_cov: Taget coverage patterns
        :return: Posteriors of coverage patterns shape: (4, 5, n)
        """
        # prevent overflow when indexing into phi_stored
        target_cov[target_cov > 990] = 990
        len_b = self.priors.phi_stored.shape[0]
        len_g = self.priors.phi_stored.shape[1]
        # how many patterns of coverage are input
        n = target_cov.shape[0]
        # repeat the priors for all input patterns
        coverage_post = np.repeat(self.priors.priors[:, np.newaxis], repeats=n, axis=1)
        # init empty phi, since we will use the stored version
        phis = np.full(n, 1.0)
        # calculate the posteriors for all coverage patterns
        for j in range(len_g):
            if j > 0:
                phis.fill(1.0)
            for i in range(len_b):
                phis *= self.priors.phi_stored[i, j, target_cov[:, i]]
            for h in range(4):
                coverage_post[h, :, j] *= phis
        # sum_b is Z_i(D) (Normalizing factor)
        # for the posterior (the likelihood of the new data at the position).
        for _h in range(4):
            sum_b = np.sum(coverage_post[_h, :, :], axis=1)
            sum_b[sum_b < 1e-300] = 1e-300  # prevent the next line to divide by zero
            coverage_post[_h, :, :] /= sum_b[:, np.newaxis]
        return coverage_post



    def calc_score(self, scores: NDArray, pos_posterior: NDArray) -> tuple[NDArray, NDArray]:
        """
        Uses the current posterior probabilities to find the benefit score of a given position.

        :param scores: Array of scores to update
        :param pos_posterior: posterior probabilites to calculate the scores
        :return: Updated arrays of scores and entropies
        """
        len_s = len(scores)
        # Shannon's entropy of the current posterior distribution.
        logs = np.log(pos_posterior, where=pos_posterior > 0.0)
        logs[logs == np.nan] = 0
        entropy = np.sum(-pos_posterior * logs, axis=1)
        # Expected entropy after a new batch of reads
        new_entropy = np.zeros(len_s)
        # probability of observing a new base of each type at the considered position
        observation_probabilities = np.zeros(len_s)
        # posterior probabilities after reading a certain base i
        new_post = np.zeros((len_s, self.priors.len_g))
        for i in range(self.priors.len_b):  # new read base
            np.multiply(pos_posterior, self.priors.phi[i], out=new_post)
            np.sum(new_post, axis=1, out=observation_probabilities)
            observation_probabilities[observation_probabilities == 0] = 1e-300
            new_post /= observation_probabilities[:, np.newaxis]
            np.log(new_post, where=new_post > 0.0, out=logs)
            logs[logs == np.nan] = 0
            for j in range(self.priors.len_g):  # genotype
                new_entropy -= observation_probabilities * new_post[:, j] * logs[:, j]
        scores[:] = entropy - new_entropy
        return scores, entropy


    @staticmethod
    def merge_benefit(contigs) -> tuple[NDArray, NDArray]:
        """
        :return: Concatenated array of read starting sites across all contigs
        """
        c_benefits = [c.additional_benefit for c in contigs.values()]
        c_smu = [c.smu for c in contigs.values()]
        return np.concatenate(c_benefits), np.concatenate(c_smu)




    @staticmethod
    def find_strat_thread(benefit: NDArray, smu: NDArray, fhat: NDArray, time_cost: int) -> tuple[NDArray, float]:
        """
        Finding approximate decision strategy from the current read benefits

        :param benefit: Additional benefit at each position
        :param smu: S_mu at each position
        :param fhat: Read starting probability at each position
        :param time_cost: Additional time cost of sequencing
        :return: Tuple of strategy and threshold of acceptance
        """
        # take downsampling into consideration
        window = 100
        alpha = 300 // window
        rho = 300 // window
        mu = 400 // window
        tc = time_cost // window
        # group benefit into bins of similar values
        # using binary exponent
        benefit_flat = benefit.flatten('F')
        benefit_nz_ind = np.nonzero(benefit_flat)
        benefit_flat_nz = benefit_flat[benefit_nz_ind]
        # to make binary exponents work, normalise benefit values
        normaliser = np.max(benefit_flat_nz)
        benefit_flat_norm = benefit_flat_nz / normaliser
        mantissa, benefit_exponents = np.frexp(benefit_flat_norm)
        # count how often each exponent is present
        # absolute value because counting positive integers is quicker
        benefit_exponents_pos = np.abs(benefit_exponents)
        # multi-thread counting of exponents
        exponent_arrays = np.array_split(benefit_exponents_pos, 12)
        with TPexe(max_workers=12) as executor:
            exponent_counts = executor.map(np.bincount, exponent_arrays)
        exponent_counts = list(exponent_counts)
        # aggregate results from threads
        # target array needs to have the largest shape of the thread results
        max_exp = np.max([e.shape[0] for e in exponent_counts])
        bincounts = np.zeros(shape=max_exp, dtype='int')
        # sum up results from individual threads
        for exp in exponent_counts:
            bincounts[0:exp.shape[0]] += exp

        exponents_unique = np.nonzero(bincounts)[0]  # filter empty bins
        counts = bincounts[exponents_unique]  # counts of the existing benefit exponents
        # perform weighted bincount in multiple threads
        exponent_arrays = np.array_split(benefit_exponents_pos, 12)
        fhat_arrays = np.array_split(fhat.flatten('F')[benefit_nz_ind], 12)
        arguments = zip(exponent_arrays, fhat_arrays)

        with TPexe(max_workers=12) as executor:
            fgs = executor.map(binc, arguments)
        fgs = list(fgs)
        # aggregates results of threads
        max_fg = np.max([f.shape[0] for f in fgs])
        f_grid = np.zeros(shape=max_fg, dtype='float')
        # aggregate results
        for fg in fgs:
            f_grid[0: fg.shape[0]] += fg

        f_grid = f_grid[exponents_unique]  # filter empty bins
        f_grid_mean = f_grid / counts  # mean fhat for exponent bin
        # use exponents to rebuild benefit values
        benefit_bin = np.power(2.0, -exponents_unique) * normaliser
        # average benefit of strategy in the case that all fragments are rejected
        ubar0 = np.sum(fhat * smu)
        tbar0 = alpha + rho + mu
        # cumsum of the benefit (bins multiplied by how many sites are in the bin)
        # weighted by probability of a read coming from that bin
        cs_u = np.cumsum(benefit_bin * f_grid_mean * counts) + ubar0
        cs_t = np.cumsum(tc * counts * f_grid_mean) + tbar0
        peak = cs_u / cs_t
        strat_size = np.argmax(peak) + 1
        # plt.plot(cs_u)
        # plt.plot(cs_t)
        # plt.plot(peak)
        # plt.show()

        # calculate threshold exponent and where values are geq
        try:
            threshold = benefit_bin[strat_size]
        except IndexError:
            threshold = benefit_bin[-1]

        strat = np.where(benefit >= threshold, True, False)
        return strat, threshold







class CoverageConverter:

    def __init__(self, qt: int = 0):
        """
        Class that handles the conversion of read mappings to coverage counts of nucleotides

        :param qt: Minimum quality threshold to count observations
        """
        # set up translation dict for bases
        transDict = {'A': '0', 'C': '1', 'G': '2', 'T': '3'}
        self.base2int = str.maketrans(transDict)
        # and a translation dict for the cigar ops
        transDict_cig = {'M': '6', 'D': '7', 'I': '8', 'S': '9'}
        self.cig2int = str.maketrans(transDict_cig)
        # compile regex for cigar ops
        self.cigar_regex = re.compile(r"(\d+)([MIDNSHP=XB])")
        # quality threshold
        self.qt = qt



    def convert_records(
            self,
            paf_dict: paf_dict_type,
            seqs: dict[str, str],
            quals: dict[str, str]) -> defaultdict[Any, list]:
        """
        Convert mappings to coverage counts

        :param paf_dict: Dict of mappings
        :param seqs: Dict of read sequences
        :param quals: Dict of read qualities
        :return: Defaultdict of lists of coverage counts per target seq
        """
        # container to collect increments per contig
        increments = defaultdict(list)
        # loop through all reads
        read_ids = list(paf_dict.keys())
        for rid in read_ids:
            # grab paf record
            rec = paf_dict[rid]
            if len(rec) > 1:
                rec = Paf.choose_best_mapper(rec)[0]
            else:
                rec = rec[0]

            # handle strands
            if rec.rev:
                seq = reverse_complement(seqs[rec.qname])
                qual = quals[rec.qname][::-1]
                start = rec.qlen - rec.qend
                end = rec.qlen - rec.qstart
            else:
                seq = seqs[rec.qname]
                qual = quals[rec.qname]
                start = rec.qstart
                end = rec.qend

            # get array of counts for each base and indels
            query_arr, qual_arr = self._parse_cigar(
                cigar_string=rec.cigar,
                read_string=seq,
                qual_string=qual,
                start=start,
                end=end
            )

            # make sure the arrays are the correct length
            # range of coverage of the read
            start = min(rec.tstart, rec.tend)
            end = max(rec.tstart, rec.tend)
            assert (end - start) == query_arr.shape[0]
            assert (end - start) == qual_arr.shape[0]
            # eliminate low qual coverage
            addition = np.ones(shape=query_arr.shape[0])
            addition[np.where(qual_arr < self.qt)] = 0
            # collect increments
            increments[rec.tname].append((start, end, query_arr, addition))
        return increments




    def _parse_cigar(
            self,
            cigar_string: str,
            read_string: str,
            qual_string: str,
            start: int,
            end: int) -> tuple[NDArray, NDArray]:
        """
        Parse cigar string to integer base observations

        :param cigar_string: A CIGAR string
        :param read_string: Read that the CIGAR string describes
        :param qual_string: string of base qualities (phred)
        :param start: starting position of the aligned bit on the read
        :param end: end position of the aligned bit on the read
        :return: tuple of integerised base and quality arrays of query at each pos
        """
        # translate sequence and qualities to integers
        read_integer = read_string.translate(self.base2int)
        int_seq = np.frombuffer(read_integer.encode(), 'u1') - ord('0')
        # qualities are simply transformed to ascii with frombuffer
        # then subtract 33 for the phred score offset
        int_qual = np.frombuffer(qual_string.encode('ascii'), dtype=np.uint8) - 33
        # split cigar into tuples of length, opcode
        parts = self.cigar_regex.findall(cigar_string)
        # cast into lists
        lengths, ops = zip(*parts)
        # convert to array
        lengths_arr = np.array(lengths, dtype=np.uint32)
        # for the opcodes, first put them into a string and translate to integers
        ops_seq = ''.join(ops)
        ops_tr = ops_seq.translate(self.cig2int)
        ops_arr = np.frombuffer(ops_tr.encode(), 'u1') - ord('0')
        # create an array that's the same length of rec.alignment_block_length
        # this consists of repeated opcodes
        cig_rep = np.repeat(ops_arr, lengths_arr)
        # all positions that are not insertions
        notins = np.where(cig_rep != 8)
        notdel = np.where(cig_rep != 7)
        cig_rep_qual = np.copy(cig_rep)
        # put the read sequence into the alignment block where it's not D
        cig_rep[notdel] = int_seq[start: end]
        # then grab the nuc that are not inserted
        query_arr = cig_rep[notins]
        # perform basically the same with the quality scores
        cig_rep_qual[notdel] = int_qual[start: end]
        qual_arr = cig_rep_qual[notins]
        # transform D to 4 in the sequence, to 20 in the qual
        qual_arr[query_arr == 7] = 20
        query_arr[query_arr == 7] = 4
        return query_arr, qual_arr











import logging
import subprocess
import os
import time
import glob
import inspect
from pathlib import Path
from datetime import datetime

import rtoml
from minknow_api.manager import Manager, FlowCellPosition
from minknow_api import __version__ as minknow_api_version

from .utils import search_running_process




class Sequencer:

    def __init__(self, position: FlowCellPosition = None):
        """
        class that represents a connected sequencing device

        :param position: FlowCellPosition instance from minknow_api
        """
        self.position = position

        if self.position:
            self._grab_output_dir()
            self._grab_device_type()
        else:   # testing pass-through
            self.channels = set()
            self.out_path = '../data'
            self.device_type = 'min'
            if not os.path.exists(f'{self.out_path}/fastq_pass'):
                os.mkdir(f'{self.out_path}/fastq_pass')




    def _grab_output_dir(self) -> None:
        """
        Capture the output directory of MinKNOW,
        i.e. where fastq files are deposited during sequencing
        :return:
        """
        # connect to the device and navigate api to get output path
        device_connection = self.position.connect()
        current_run = device_connection.protocol.get_current_protocol_run()
        run_id = current_run.run_id
        logging.info(f"connected to run_id: {run_id}")
        self.out_path = str(current_run.output_path)
        logging.info(f"grabbing Minknow's output path: \n{self.out_path}\n")

        # avoid error of directory not existing yet
        if not Path(self.out_path).exists():
            dir_exists = False
            retry_counter = 0
            while not dir_exists:
                if not Path(self.out_path).exists():
                    logging.info("output path does not exist (yet), waiting for 10s")
                    time.sleep(10)
                    retry_counter += 1
                else:
                    dir_exists = True
                if retry_counter > 10:
                    raise RuntimeError(f"Output path {self.out_path} not available after 10 retries. Exiting.")



    def _grab_device_type(self) -> None:
        """
        Get info on the type of device that the sequencing is running on.
        Can be either minion, gridion, promethion, p2_solo
        Currently not used for anything

        :return:
        """
        pro_device_types = {'P2_SOLO', 'PROMETHION'}
        min_device_types = {'MINION'}
        dt = self.position.device_type
        if dt in pro_device_types:
            self.device_type = 'pro'
        elif dt in min_device_types:
            self.device_type = 'min'
        else:
            logging.info(f'WARNING: device type {dt} not recognized. Using MINION flowcell layout as fallback')
            self.device_type = 'min'



    def grab_channels(self, run_name: str) -> None:
        """
        Wait until we find a channels TOML spec that contains the channel
        numbers assigned to the BOSS region on the flowcell.
        Result is the set of channel IDs from which data is used or empty set

        :param run_name: Experiment name from config TOML
        :return:
        """
        self.channels_toml = f'{self.out_path}/channels.toml'
        logging.info(f'looking for channels specification at : {self.channels_toml}')
        channels_found = False
        channels = []
        retry_counter = 0
        while not channels_found:
            if not os.path.isfile(self.channels_toml):
                logging.info("channels file does not exist (yet), waiting for 30s")
                time.sleep(30)
                retry_counter += 1
            else:
                channels = self._parse_channels_toml(run_name=run_name)
                channels_found = True
            if retry_counter > 5:
                raise RuntimeError(f"Channels file {self.channels_toml} not available after 5 retries. Exiting.")
        # channels successfully found
        self.channels = channels



    def _parse_channels_toml(self, run_name: str) -> set:
        """
        Look into the channels toml that readfish writes
        This toml contains lists of channels assigned to each region
        Grab the channel numbers for the BOSS region

        :param run_name: experiment name of BOSS region
        :return: Set of channel numbers from which to consider data
        """
        toml_dict = rtoml.load(Path(self.channels_toml))
        # if there is only one condition, we return empty set
        # that way we can skip the regex when scanning new data
        if len(toml_dict["conditions"]) == 1:
            logging.info('Only one condition. Using all channels!')
            return set()

        # otherwise find the corresponding condition
        correct_key = ''
        for key in toml_dict["conditions"].keys():
            name = toml_dict["conditions"][key]["name"]
            if name == run_name:
                correct_key = key
                break

        if not correct_key:
            raise ValueError(f"Experiment name {run_name} in .toml not found in channel-specification toml.")

        selected_channels = set(toml_dict["conditions"][correct_key]["channels"])
        logging.info(f"found channels specification: Using {len(selected_channels)} channels.")
        return selected_channels




class LiveRun:


    @staticmethod
    def connect_sequencer(device: str, host: str = 'localhost', port: int = 9502) -> Sequencer:
        """
        Connect to the running sequencer to get the path to its output directory

        :param device: Device name on sequencing machine
        :param host: Host to connect to
        :param port: Possibility to overwrite default port
        :return: Sequencer object that holds info about the flowcellposition
        """
        LiveRun._check_minknow_api_version()
        # try connecting to the sequencing device
        try:
            flowcellposition = LiveRun._grab_target_device(device=device, host=host, port=port)
            sequencer = Sequencer(position=flowcellposition)
        except:
            raise ValueError(f"Error: issue connecting to target device {device}, at host {host} and port {port}. \n"
                             f"Please make sure to supply correct name of sequencing position in MinKNOW")
        return sequencer


    @staticmethod
    def _check_minknow_api_version() -> None:
        """
        Check compatibility of the available minknow_api version. Hard exits if not compatible
        :return:
        """
        logging.info(f"minknow API Version {minknow_api_version}")
        # minknow_api.manager supplies Manager (wrapper around MinKNOW's Manager gRPC)
        if not minknow_api_version.startswith("6"):
            raise NotImplementedError("Unsupported version of minknow_api. MinKnow <6 is not supported.")



    @staticmethod
    def _grab_target_device(device: str, host: str = 'localhost', port: int = 9502) -> FlowCellPosition:
        """
        Get the flowcell position that runs the sequencing process

        :param device: device name of the 'position' in the sequencing machine
        :param host: hostname to connect to for MinKNOW
        :param port: override default port to connect to
        :return: instance of FlowCellPosition from minknow_api
        """
        # Find a list of currently available sequencing positions.
        manager = Manager(host=host, port=int(port))
        positions = list(manager.flow_cell_positions())
        pos_dict = {pos.name: pos for pos in positions}
        # index into the dict of available devices
        target_device = pos_dict[device]
        return target_device



    @staticmethod
    def scan_dir(fastq_pass: str, processed_files: set) -> list[str]:
        """
        Periodically scan Minknow's output dir and
        return list of all NEW files

        :param fastq_pass: Path of MinKNOW's output
        :param processed_files: Set of already processed files
        :return: List of new, previously unprocessed files
        """
        patterns = ["*.fq.gz", "*.fastq.gz", "*.fastq.gzip", "*.fq.gzip", "*.fastq", "*.fq"]
        all_fq = set()
        for p in patterns:
            all_fq.update(glob.glob(f'{fastq_pass}/{p}'))
        # which files have we not seen before?
        new_fq = all_fq.difference(processed_files)
        logging.info(f"found {len(new_fq)} new fq files")
        new_fq_list = [f for f in new_fq]
        return new_fq_list



    @staticmethod
    def launch_readfish(toml: str, device: str, name: str) -> str:
        """
        Wrapper to launch readfish into the background

        :param toml: TOML config for readfish
        :param device: Flowcell position
        :param name: Name of BOSS experiment
        :return: name of the readfish log file
        """
        if device == "TEST":  # exit for testing purposes
            return ''
        if not Path(toml).exists():
            raise FileNotFoundError("Specified readfish toml does not exist.")
        if search_running_process('readfish'):
            raise RuntimeError("readfish is already running. Please kill the process before starting a new one.")
        # find the script to launch readfish
        module_path = inspect.getfile(LiveRun)
        logging.info(module_path)
        script_path = Path(module_path).parent / "readfish_boss.py"
        if not script_path.is_file():
            raise FileNotFoundError("readfish_boss.py not found. Something went wrong..")
        stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        readfish_log = f'{stamp}_readfish.log'
        readfish_comm = f'python {script_path} {toml} {device} {name} >{readfish_log} 2>&1'
        logging.info(readfish_comm)
        # launch readfish into the background
        logging.info("Launching readfish")
        subprocess.Popen(readfish_comm, shell=True)
        return readfish_log


